"""
Base classes for GEO types
"""

__all__ = [
    "GeoNestedArrayNumbaType",
    "HeavyDBGeoNestedArray",
    "GeoNestedArray",
    "heavydb_geo_fromCoords_vec",
    "heavydb_geo_fromCoords_vec2",
    "heavydb_geo_fromCoords_vec3",
    "heavydb_geo_toCoords_vec",
    "heavydb_geo_toCoords_vec2",
    "heavydb_geo_toCoords_vec3",
]

import operator

from llvmlite import ir
from numba.core import cgutils, extending
from numba.core import types as nb_types

from rbc import typesystem
from rbc.external import external

from .allocator import allocate_varlen_buffer
from .metatype import HeavyDBMetaType
from .utils import as_voidptr, get_alloca

i1 = ir.IntType(1)
i8 = ir.IntType(8)
i8p = i8.as_pointer()
i8pp = i8p.as_pointer()
i32 = ir.IntType(32)
i64 = ir.IntType(64)
i64p = i64.as_pointer()
void = ir.VoidType()


class GeoNestedArray(metaclass=HeavyDBMetaType):
    """
    Base class for GEO types.

    .. note::
        Geo columns should inherit from ``ColumnFlatBuffer`` in
        ``column_flatbuffer.py``.

    .. code-block:: c

        {
            int8_t* flatbuffer_;
            int64_t index_[4];
            int64_t n_;
        }
    """

    def __len__(self) -> int:
        """
        """

    def size(self) -> int:
        """
        """

    def is_null(self) -> bool:
        """
        Return wether the Geo type is ``NULL``
        """


class GeoNestedArrayNumbaType(nb_types.Type):
    def __init__(self, name):
        super().__init__(name)
        self.base_type = self.__typesystem_type__._params["base_type"]
        self.item_type = self.__typesystem_type__._params["item_type"]


class HeavyDBGeoNestedArray(typesystem.Type):
    """Typesystem type class for HeavyDB buffer structures."""

    @property
    def numba_type(self):
        return GeoNestedArrayNumbaType

    @property
    def type_name(self):
        raise NotImplementedError()

    @property
    def item_type(self):
        raise NotImplementedError()

    def postprocess_type(self):
        return self.params(shorttypename=self.type_name)

    @property
    def custom_params(self):
        return {
            "NumbaType": self.numba_type,
            "name": self.type_name,
            "base_type": self.type_name,
            "item_type": self.item_type,
        }

    def tonumba(self, bool_is_int8=None):
        flatbuffer_t = typesystem.Type.fromstring("int8_t* flatbuffer_")
        index_t = typesystem.Type.fromstring("int64_t index_")
        n_t = typesystem.Type.fromstring("int64_t n_")
        typ = typesystem.Type(
            flatbuffer_t,
            index_t,
            index_t,
            index_t,
            index_t,
            n_t,
        )
        typ.params(other=None, **self.custom_params)
        return typ.tonumba(bool_is_int8=True)


@extending.intrinsic
def get_c_ptr(typingctx, sz, typ):
    t = typ.dtype
    assert isinstance(typ, nb_types.DTypeSpec)
    sig = nb_types.CPointer(t)(sz, typ)

    def codegen(context, builder, sig, args):
        [sz, _] = args
        # TODO: replace i64(8) by i64(sizeof(typ))
        ptr = allocate_varlen_buffer(builder, sz, i64(8))
        typ = sig.args[1].dtype
        llty = context.get_value_type(typ)
        llty_p = llty.as_pointer()
        return builder.bitcast(ptr, llty_p)

    return sig, codegen


@extending.intrinsic
def heavydb_geo_getitem_(typingctx, geo, index):
    retty = typesystem.Type.fromstring(geo.item_type).tonumba()
    sig = retty(geo, index)

    def codegen(context, builder, sig, args):
        geo, index = args
        fnty = ir.FunctionType(void, [i8p, i64, i8p])
        getItem = cgutils.get_or_insert_function(
            builder.module, fnty, f"{sig.args[0].base_type}_getItem"
        )
        geo_ptr = builder.bitcast(
            context.make_helper(builder, sig.args[0], value=geo)._getpointer(), i8p
        )

        # Alloca ItemType
        fa = context.make_helper(builder, sig.return_type)

        # ItemType -> void*
        result_ptr = builder.bitcast(fa._getpointer(), i8p)

        # call func
        builder.call(getItem, [geo_ptr, index, result_ptr])

        # convert void* -> ItemType
        point_type = fa._get_be_type(fa._datamodel)
        return builder.load(builder.bitcast(result_ptr, point_type.as_pointer()))

    return sig, codegen


@extending.overload(len)
@extending.overload_method(GeoNestedArrayNumbaType, "size")
def heavydb_geo_len(geo):
    if isinstance(geo, GeoNestedArrayNumbaType):
        base_type = geo.base_type
        size_ = external(f"int64_t {base_type}_size(int8_t*)")

        def impl(geo):
            return size_(as_voidptr(get_alloca(geo)))

        return impl


@extending.overload_method(GeoNestedArrayNumbaType, "is_null")
def heavydb_geo_isNull(geo):
    if isinstance(geo, GeoNestedArrayNumbaType):
        base_type = geo.base_type
        isNull = external(f"bool {base_type}_isNull(int8_t*)")

        def impl(geo):
            return isNull(as_voidptr(get_alloca(geo)))

        return impl


@extending.overload(operator.getitem)
@extending.overload_method(GeoNestedArrayNumbaType, "get_item")
def heavydb_geo_getitem(geo, index):
    if isinstance(geo, GeoNestedArrayNumbaType):

        def impl(geo, index):
            return heavydb_geo_getitem_(geo, index)

        return impl


def heavydb_geo_fromCoords_vec(geo, lst):
    base_type = geo.base_type
    fromCoords = external(
        f"void {base_type}_fromCoords(int8_t*, double*, int64_t size)"
    )

    def impl(geo, lst):
        sz = len(lst)

        f64p = get_c_ptr(sz, nb_types.double)
        for i in range(sz):
            f64p[i] = lst[i]

        return fromCoords(as_voidptr(get_alloca(geo)), f64p, sz)

    return impl


def heavydb_geo_fromCoords_vec2(geo, lst):
    base_type = geo.base_type
    fromCoords = external(
        f"void {base_type}_fromCoords(int8_t*, double*, int64_t*, int64_t)"
    )

    def impl(geo, lst):
        dim0 = len(lst)
        geo_ptr = as_voidptr(get_alloca(geo))

        indices = get_c_ptr(dim0, nb_types.int64)
        n_elems = 0
        for i in range(dim0):
            n_elems += len(lst[i])
            indices[i] = len(lst[i])

        data = get_c_ptr(n_elems, nb_types.double)
        idx = 0
        for i in range(dim0):
            dim1 = len(lst[i])
            for j in range(dim1):
                data[idx] = lst[i][j]
                idx += 1

        return fromCoords(geo_ptr, data, indices, dim0)

    return impl


def heavydb_geo_fromCoords_vec3(geo, lst):
    base_type = geo.base_type
    fromCoords = external(
        f"void {base_type}_fromCoords(int8_t*, double*, int64_t*, int64_t*, int64_t, int64_t)"
    )

    def impl(geo, lst):
        geo_ptr = as_voidptr(get_alloca(geo))

        indices_i = []
        indices_j = []
        flatten = []

        dim0 = len(lst)
        for i in range(dim0):
            cnt = 0
            dim1 = len(lst[i])
            for j in range(dim1):
                dim2 = len(lst[i][j])
                cnt += dim2
                indices_j.append(dim2)
                for k in range(dim2):
                    flatten.append(lst[i][j][k])
            indices_i.append(cnt)

        data = get_c_ptr(len(flatten), nb_types.double)
        for i, e in enumerate(flatten):
            data[i] = e

        indices_i_ = get_c_ptr(len(indices_i), nb_types.int64)
        for i, e in enumerate(indices_i):
            indices_i_[i] = e

        indices_j_ = get_c_ptr(len(indices_j), nb_types.int64)
        for j, e in enumerate(indices_j):
            indices_j_[j] = e

        return fromCoords(
            geo_ptr, data, indices_i_, indices_j_, len(indices_i), len(indices_j)
        )

    return impl


def heavydb_geo_toCoords_vec(geo):
    base_type = geo.base_type
    get_value = external(f"double {base_type}_toCoords_get_value(int8_t*, int64_t)")

    def impl(geo):
        geo_ptr = as_voidptr(get_alloca(geo))

        lst = []
        n_coords = geo.n_coords()
        for i in range(n_coords):
            lst.append(get_value(geo_ptr, i))
        return lst

    return impl


def heavydb_geo_toCoords_vec2(geo):
    base_type = geo.base_type
    get_value = external(
        f"double {base_type}_toCoords_get_value(int8_t*, int64_t, int64_t)"
    )

    def impl(geo):
        lst = []
        geo_ptr = as_voidptr(get_alloca(geo))
        n_rings = geo.n_rings()
        for i in range(n_rings):
            n_coords = geo[i].n_coords()
            inner = []
            for j in range(n_coords):
                inner.append(get_value(geo_ptr, i, j))
            lst.append(inner)
        return lst

    return impl


def heavydb_geo_toCoords_vec3(geo):
    raise NotImplementedError()




"""Implement Heavydb Column<Array<T>> type support

Heavydb Column<Array<T>> type is the type of input/output column arguments in
UDTFs.
"""

__all__ = ['HeavyDBColumnListArrayType', 'ColumnListArray']


import operator
from rbc import typesystem
from .column_list import HeavyDBColumnListType, ColumnList
from numba.core import extending, cgutils
from numba.core import types as nb_types
from llvmlite import ir


i1 = ir.IntType(1)
i8 = ir.IntType(8)
i8p = i8.as_pointer()
i8pp = i8p.as_pointer()
i64 = ir.IntType(64)
i64p = i64.as_pointer()
void = ir.VoidType()


_COLUMN_PARAM_NAME = 'ColumnListArray_inner_type'


class ColumnListArray(ColumnList):
    """
    RBC ``ColumnList<Array<T>>`` type that corresponds to HeavyDB COLUMN LIST
    of arrays.

    In HeavyDB, a ColumnList<Array<T>> of type ``T`` is represented as follows:

    .. code-block:: c

        {
            int8_t* flatbuffer_;
            int64_t size_;
        }

    """


class HeavyDBColumnListArrayType(HeavyDBColumnListType):

    def postprocess_type(self):
        return self

    @property
    def numba_type(self):
        return ColumnListArrayType

    @property
    def numba_pointer_type(self):
        return ColumnListArrayPointer

    @property
    def element_type(self):
        return typesystem.Type.fromstring('int8_t')

    @property
    def custom_params(self):
        return {
            'NumbaType': self.numba_type,
            'NumbaPointerType': self.numba_pointer_type,
            _COLUMN_PARAM_NAME: self[0][0],
            'name': f'STRUCT_{self.mangling()}_{self[0][0][0][0].tostring()}'
        }


class ColumnListArrayType(nb_types.Type):
    """Numba type class for HeavyDB ColumnList<Array<T>> structures.
    """

    def __init__(self, dtype):
        array_T = self.__typesystem_type__._params.get(_COLUMN_PARAM_NAME)
        eltype = array_T.tonumba().eltype
        self.dtype = array_T.tonumba()  # struct dtype
        name = f"ColumnList<Array<{eltype}>>"
        super().__init__(name)

    @property
    def eltype(self):
        """
        Return buffer element dtype.
        """
        return self.dtype.eltype


class ColumnListArrayPointer(nb_types.Type):
    """Numba type class for pointers to HeavyDB buffer structures.

    We are not deriving from CPointer because ColumnListArrayPointer getitem is
    used to access the data stored in Buffer ptr member.
    """
    mutable = True
    return_as_first_argument = True

    def __init__(self, dtype):
        self.dtype = dtype    # struct dtype
        name = "%s[%s]*" % (type(self).__name__, dtype)
        super().__init__(name)

    @property
    def key(self):
        return self.dtype


@extending.intrinsic
def heavydb_column_list_array_nrows_(typingctx, x):
    sig = nb_types.int64(x)

    def codegen(context, builder, sig, args):
        [col] = args
        proxy = cgutils.create_struct_proxy(sig.args[0].dtype)
        clst = proxy(context, builder, value=builder.load(col))
        return clst.num_rows_

    return sig, codegen


@extending.intrinsic
def heavydb_column_list_array_ncols_(typingctx, x):
    sig = nb_types.int64(x)

    def codegen(context, builder, sig, args):
        [col] = args
        proxy = cgutils.create_struct_proxy(sig.args[0].dtype)
        clst = proxy(context, builder, value=builder.load(col))
        return clst.num_cols_

    return sig, codegen


@extending.intrinsic
def heavydb_column_list_array_getitem_(typingctx, x, i):
    T = x.dtype.eltype
    col = typesystem.Type.fromstring(f'Column<Array<{T}>>').tonumba().dtype
    sig = col(x, i)

    def codegen(context, builder, sig, args):
        ptr, index = args
        clst = cgutils.create_struct_proxy(sig.args[0].dtype)(
            context, builder, value=builder.load(ptr))
        col = cgutils.create_struct_proxy(sig.return_type)(context, builder)
        col.ptr = builder.load(builder.gep(clst.ptrs, [index]))
        col.sz = clst.num_rows_
        return col._getvalue()

    return sig, codegen


@extending.overload(operator.getitem)
def heavydb_column_list_array_getitem(x, i):
    if isinstance(x, ColumnListArrayPointer):
        def impl(x, i):
            return heavydb_column_list_array_getitem_(x, i)
        return impl


@extending.overload_attribute(ColumnListArrayPointer, 'nrows')
def get_nrows(clst):
    def impl(clst):
        return heavydb_column_list_array_nrows_(clst)
    return impl


@extending.overload_attribute(ColumnListArrayPointer, 'ncols')
def get_ncols(clst):
    def impl(clst):
        return heavydb_column_list_array_ncols_(clst)
    return impl



"""
Array API specification for utility functions.

https://data-apis.org/array-api/latest/API_specification/utility_functions.html
"""
from rbc.stdlib import Expose

__all__ = ["all", "any"]

expose = Expose(globals(), "utility_functions")


@expose.not_implemented("all")
def _array_api_all(x, *, axis=None, keepdims=False):
    """
    Tests whether all input array elements evaluate to `True` along a specified axis.
    """
    pass


@expose.not_implemented("any")
def _array_api_any(x, *, axis=None, keepdims=False):
    """
    Tests whether any input array element evaluates to `True` along a specified axis.
    """
    pass



"""
Implements a common interface for pointers to structs defined in HeavyDB.
    + TableFunctionManager
    + RowFunctionManager
    + String Dictionary Proxy
"""

__all__ = ['OpaquePtrNumbaType', 'HeavyDBOpaquePtr']


from numba.core import types as nb_types
from numba.core import extending, datamodel

from rbc import typesystem


class OpaquePtrNumbaType(nb_types.CPointer):
    """
    """
    def __init__(self) -> None:
        super().__init__(dtype=nb_types.int8)


class HeavyDBOpaquePtr(typesystem.Type):

    def register_datamodel(self):
        name = self.type_name
        data_model = type(name+'Model',
                          (datamodel.models.PointerModel,),
                          dict())
        extending.register_model(self.numba_type)(data_model)

    def postprocess_type(self):
        self.register_datamodel()
        return super().postprocess_type()

    @property
    def numba_type(self):
        raise NotImplementedError()

    @property
    def type_name(self):
        raise NotImplementedError()

    def tonumba(self, bool_is_int8=None):
        return self.numba_type()



import pytest
import re
from rbc.tests import heavydb_fixture
from rbc.errors import HeavyDBServerError


@pytest.fixture(scope='module')
def heavydb():

    for o in heavydb_fixture(globals(), minimal_version=(6, 1), suffices=['']):
        yield o


@pytest.mark.parametrize("func", ['dbscan', 'kmeans'])
def test_mlpack(heavydb, func):
    heavydb.require_version(
        (6, 1),
        'Requires heavydb-internal PR 5430 and heavydb built with -DENABLE_MLPACK')

    extra_args = dict(dbscan='cast(1 as float), 1, \'mlpack\'',
                      kmeans='1, 1, \'default\', \'mlpack\'')[func]
    query = (f'select * from table({func}(cursor(select cast(rowid as int), f8, f8, f8 '
             f'from {heavydb.table_name}), {extra_args}))')

    try:
        _, result = heavydb.sql_execute(query)
    except HeavyDBServerError as msg:
        m = re.match(r'.*Cannot find (mlpack|DEFAULT) ML library to support', msg.args[0])
        if m:
            pytest.skip(f'test requires heavydb server with MLPACK support: {msg}')
        # heavydb from conda-forge is built without MLPACK
        m = re.match(r".*Undefined function call '(kmeans|dbscan)' in SQL statement.*",
                     msg.args[0])
        if m:
            pytest.skip(f'test requires heavydb server with MLPACK support: {msg}')
        raise

    result = list(result)

    expected = dict(
        dbscan=[(0, 0), (1, 1), (2, 2), (3, 3), (4, 4)],
        kmeans=[(0, 0), (1, 0), (2, 0), (3, 0), (4, 0)]
    )[func]

    assert result == expected



"""Implement Heavydb Column<GeoPoint> type support

Heavydb Column<GeoPoint> type is the type of input/output column arguments in
UDTFs.
"""

__all__ = [
    "HeavyDBOutputColumnGeoPointType",
    "HeavyDBColumnGeoPointType",
    "ColumnGeoPoint",
]

import operator

from llvmlite import ir
from numba.core import cgutils, extending

from rbc import typesystem

from . import geopoint
from .column_flatbuffer import (ColumnFlatBuffer, ColumnFlatBufferPointer,
                                ColumnFlatBufferType,
                                HeavyDBColumnFlatBufferType,
                                HeavyDBOutputColumnFlatBufferType)
from .metatype import HeavyDBMetaType

i1 = ir.IntType(1)
i8 = ir.IntType(8)
i8p = i8.as_pointer()
i8pp = i8p.as_pointer()
i32 = ir.IntType(32)
i64 = ir.IntType(64)
i64p = i64.as_pointer()
void = ir.VoidType()


class ColumnGeoPoint(ColumnFlatBuffer, metaclass=HeavyDBMetaType):
    """
    RBC ``Column<GeoPoint>`` type that corresponds to HeavyDB
    ``COLUMN<POINT>``

    In HeavyDB, a Column of type ``GeoPoint`` is represented as follows:

    .. code-block:: c

        {
            int8_t* flatbuffer;
            int64_t size;
        }

    """

    def __getitem__(self, index: int) -> 'geopoint.GeoPoint':
        """
        Return ``self[index]``

        .. note::
            Only available on ``CPU``
        """

    def get_item(self, index: int) -> 'geopoint.GeoPoint':
        """
        Return ``self[index]``

        .. note::
            Only available on ``CPU``
        """

    def set_item(self, index: int, buf: 'geopoint.GeoPoint') -> None:
        """
        Set line from a buffer of point coordindates

        .. note::
            Only available on ``CPU``
        """

    def __setitem__(self, index: int, buf: 'geopoint.GeoPoint') -> None:
        """
        Set line from a buffer of point coordindates

        .. note::
            Only available on ``CPU``
        """


class HeavyDBColumnGeoPointType(HeavyDBColumnFlatBufferType):
    """ """

    @property
    def numba_type(self):
        return ColumnGeoPointType

    @property
    def numba_pointer_type(self):
        return ColumnGeoPointPointer

    @property
    def type_name(self):
        return "GeoPoint"


class HeavyDBOutputColumnGeoPointType(
    HeavyDBColumnGeoPointType, HeavyDBOutputColumnFlatBufferType
):
    """Heavydb OutputColumn type for RBC typesystem.

    OutputColumn<GeoPoint> is the same as Column<GeoPoint> but introduced to
    distinguish the input and output arguments of UDTFs.
    """


class ColumnGeoPointType(ColumnFlatBufferType):
    pass


class ColumnGeoPointPointer(ColumnFlatBufferPointer):
    pass


@extending.intrinsic
def heavydb_column_getitem_(typingctx, col, index):
    Point2D = typesystem.Type.fromstring("Point2D").tonumba()
    # oddly enough, a Column<GeoPoint> returns a Point2D
    sig = Point2D(col, index)

    def codegen(context, builder, sig, args):
        col, index = args
        fnty = ir.FunctionType(void, [i8p, i64, i32, i8p])
        getItem = cgutils.get_or_insert_function(
            builder.module, fnty, "ColumnGeoPoint_getItem"
        )
        output_srid = i32(0)
        col_ptr = builder.bitcast(col, i8p)

        # Alloca Point2D
        fa = context.make_helper(builder, sig.return_type)

        # Point2D -> void*
        result_ptr = builder.bitcast(fa._getpointer(), i8p)

        # call func
        builder.call(getItem, [col_ptr, index, output_srid, result_ptr])

        # convert void* -> Point2D
        point_type = fa._get_be_type(fa._datamodel)
        return builder.load(builder.bitcast(result_ptr, point_type.as_pointer()))

    return sig, codegen


@extending.overload(operator.getitem)
@extending.overload_method(ColumnGeoPointPointer, "get_item")
def heavydb_column_getitem(col, index):
    if isinstance(col, ColumnGeoPointPointer):

        def impl(col, index):
            return heavydb_column_getitem_(col, index)

        return impl



import os
os.environ["NUMBA_CAPTURED_ERRORS"] = "new_style"
os.environ["NUMBA_LOOP_VECTORIZE"] = "0"
os.environ["NUMBA_SLP_VECTORIZE"] = "0"

# Expose a temporary prototype. It will be replaced by proper
# implementation soon.
from .remotejit import RemoteJIT  # noqa: F401, E402
from .heavydb import RemoteHeavyDB  # noqa: F401, E402

from ._version import get_versions  # noqa: E402
__version__ = get_versions()['version']
del get_versions



"""RBC YearMonthTimeInterval type."""
import operator
from rbc import typesystem
from rbc.heavydb import HeavyDBMetaType
from rbc.external import external
from rbc.targetinfo import TargetInfo
from numba.core import extending, cgutils, datamodel
from numba.core import types as nb_types
from .timestamp import TimestampNumbaType, Timestamp

__all__ = [
    "HeavyDBYearMonthTimeIntervalType",
    "YearMonthTimeInterval",
    "YearMonthTimeIntervalNumbaType",
]


daYEAR = 0
daQUARTER = 1
daMONTH = 2
daDAY = 3
daHOUR = 4
daMINUTE = 5
daSECOND = 6
daMILLENNIUM = 7
daCENTURY = 8
daDECADE = 9
daMILLISECOND = 10
daMICROSECOND = 11
daNANOSECOND = 12
daWEEK = 13
daQUARTERDAY = 14
daWEEKDAY = 15
daDAYOFYEAR = 16
daINVALID = 17


class YearMonthTimeInterval(metaclass=HeavyDBMetaType):
    """
        RBC ``YearMonthTimeInterval`` type.

        .. code-block:: c

            struct YearMonthTimeInterval {
                int64_t timeval;
            };

    """

    def __init__(self, timeval: int) -> 'YearMonthTimeInterval':
        pass

    def numStepsBetween(self, start: Timestamp, stop: Timestamp) -> int:
        """
        Return a new  ``Timestamp`` with ``time`` truncated to microseconds
        """
        pass

    def __eq__(self, other: "YearMonthTimeInterval") -> bool:
        """ """
        pass

    def __ne__(self, other: "YearMonthTimeInterval") -> bool:
        """ """
        pass

    def __mul__(self, other: int) -> bool:
        """ """
        pass

    def __add__(self, other: Timestamp) -> bool:
        """ """
        pass


class YearMonthTimeIntervalNumbaType(nb_types.Type):
    def __init__(self):
        super().__init__(name="YearMonthTimeInterval")


@extending.register_model(YearMonthTimeIntervalNumbaType)
class YearMonthTimeIntervalModel(datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [
            ("timeval", nb_types.int64),
        ]
        datamodel.models.StructModel.__init__(self, dmm, fe_type, members)


extending.make_attribute_wrapper(YearMonthTimeIntervalNumbaType, "timeval", "timeval")


class HeavyDBYearMonthTimeIntervalType(typesystem.Type):
    """Typesystem type class for HeavyDB YearMonthTimeInterval structures."""

    @property
    def __typesystem_type__(self):
        return typesystem.Type.fromstring("{int64 timeval}").params(
            name="YearMonthTimeInterval", NumbaType=YearMonthTimeIntervalNumbaType
        )

    def tonumba(self, bool_is_int8=None):
        return YearMonthTimeIntervalNumbaType()

    def postprocess_type(self):
        return self.params(shorttypename='YearMonthTimeInterval')


@extending.type_callable(YearMonthTimeInterval)
def type_heavydb_YearMonthTimeInterval(context):
    def typer(arg):
        if isinstance(arg, nb_types.Integer):
            return typesystem.Type.fromobject("YearMonthTimeInterval").tonumba()

    return typer


@extending.lower_builtin(YearMonthTimeInterval, nb_types.Integer)
def heavydb_YearMonthTimeInterval_int_ctor(context, builder, sig, args):
    timeval = args[0]
    typ = sig.return_type
    dti = cgutils.create_struct_proxy(typ)(context, builder)
    dti.timeval = timeval
    return dti._getvalue()


@extending.overload(operator.mul)
def ol_day_time_interval_operator_mul(a, t):
    if isinstance(a, YearMonthTimeIntervalNumbaType) and isinstance(t, nb_types.Integer):

        def impl(a, t):
            return YearMonthTimeInterval(a.timeval * t)

        return impl


@extending.overload(operator.eq)
def ol_day_time_interval_operator_eq(a, b):
    if isinstance(a, YearMonthTimeIntervalNumbaType) and isinstance(
        b, YearMonthTimeIntervalNumbaType
    ):

        def impl(a, b):
            return a.timeval == b.timeval

        return impl


@extending.overload(operator.ne)
def ol_day_time_interval_operator_ne(a, b):
    if isinstance(a, YearMonthTimeIntervalNumbaType) and isinstance(
        b, YearMonthTimeIntervalNumbaType
    ):

        def impl(a, b):
            return not (a == b)

        return impl


@extending.overload(operator.add)
def ol_day_time_interval_operator_add(a, b):
    from .timestamp import Timestamp, TimestampNumbaType

    target_info = TargetInfo()
    i64_null_val = target_info.null_values["int64"]
    if target_info.is_cpu:
        DateAddHighPrecisionNullable = external(
            "int64 DateAddHighPrecisionNullable(int64, int64, int64, int32, int64)"
        )

    if isinstance(a, YearMonthTimeIntervalNumbaType) and isinstance(b, TimestampNumbaType):

        def impl(a, b):
            t = DateAddHighPrecisionNullable(
                daMONTH, a.timeval, b.time, 9, i64_null_val
            )
            return Timestamp(t)

        return impl
    elif isinstance(a, TimestampNumbaType) and isinstance(b, YearMonthTimeIntervalNumbaType):

        def impl(a, b):
            t = DateAddHighPrecisionNullable(
                daMONTH, b.timeval, a.time, 9, i64_null_val
            )
            return Timestamp(t)

        return impl


@extending.overload_method(YearMonthTimeIntervalNumbaType, "numStepsBetween")
def ol_day_time_interval_numStepBetween_method(dti, begin, end):

    if isinstance(begin, TimestampNumbaType) and isinstance(end, TimestampNumbaType):

        def impl(dti, begin, end):
            timeval = dti.timeval
            if (timeval > 0 and end.time < begin.time) or (
                timeval < 0 and end.time > begin.time
            ):
                return -1

            first = end.getYear() * 12 + end.getMonth()
            second = begin.getYear() * 12 + begin.getMonth()
            return (first - second) // timeval
        return impl



"""RBC GeoMultiLineString type that corresponds to HeavyDB type GeoMultiLineString.
"""

__all__ = ["HeavyDBGeoMultiLineStringType", "GeoMultiLineString"]


from typing import List

from numba.core import extending
from numba.core import types as nb_types

from . import geolinestring
from .geo_nested_array import (GeoNestedArray, GeoNestedArrayNumbaType,
                               HeavyDBGeoNestedArray,
                               heavydb_geo_fromCoords_vec2,
                               heavydb_geo_toCoords_vec2)


class GeoMultiLineStringNumbaType(GeoNestedArrayNumbaType):
    def __init__(self, name):
        super().__init__(name)


class HeavyDBGeoMultiLineStringType(HeavyDBGeoNestedArray):
    """Typesystem type class for HeavyDB buffer structures."""

    @property
    def numba_type(self):
        return GeoMultiLineStringNumbaType

    @property
    def type_name(self):
        return "GeoMultiLineString"

    @property
    def item_type(self):
        return "GeoLineString"


class GeoMultiLineString(GeoNestedArray):
    """
    RBC ``GeoMultiLineString`` type that corresponds to HeavyDB type
    ``MULTILINESTRING``.

    .. code-block:: c

        struct MultiLineString {
            int8_t* flatbuffer_;
            int64_t index_[4];
            int64_t n_;
        }
    """
    def __getitem__(self, index: int) -> 'geolinestring.GeoLineString':
        """
        Return the ``LINESTRING`` at given index
        """

    def get_item(self, index: int) -> 'geolinestring.GeoLineString':
        """
        Return the ``LINESTRING`` at given index
        """

    def to_coords(self) -> List[List[float]]:
        """
        .. note::
            Only available on ``CPU``
        """

    def from_coords(self, coords: List[List[float]]) -> None:
        """
        .. note::
            Only available on ``CPU``
        """


@extending.overload_method(GeoMultiLineStringNumbaType, "from_coords")
def heavydb_geomultilinestring_fromCoords(geo, lst):
    if isinstance(lst, nb_types.List):
        return heavydb_geo_fromCoords_vec2(geo, lst)


@extending.overload_method(GeoMultiLineStringNumbaType, "to_coords")
def heavydb_geomultilinestring_toCoords(geo):
    return heavydb_geo_toCoords_vec2(geo)


@extending.overload_method(GeoMultiLineStringNumbaType, "n_rings")
def heavydb_geomultilinestring_nrings(geo):
    # In GeoPolygon and GeoMultiLineString, len(geo) gives the number of rings
    def impl(geo):
        return len(geo)
    return impl



"""Implement Buffer type as a base class to HeavyDB Array and Column types.

HeavyDB Buffer represents the following structure:

  template<typename T>
  struct Buffer {
    T* ptr;
    size_t sz;
    ...
  }

that is, a structure that has at least two members where the first is
a pointer to some data type and the second is the size of the buffer.

This module implements the support for the following Python operators:

  len
  __getitem__
  __setitem__

to provide a minimal functionality for accessing and manipulating the
HeavyDB buffer objects from UDF/UDTFs.
"""


import operator

import numpy as np
from llvmlite import ir
from numba.core import cgutils, datamodel, extending, imputils, types

from rbc import typesystem
from rbc.targetinfo import TargetInfo

from .allocator import allocate_varlen_buffer
from .metatype import HeavyDBMetaType

int8_t = ir.IntType(8)
int32_t = ir.IntType(32)
int64_t = ir.IntType(64)
void_t = ir.VoidType()
fp32 = ir.FloatType()
fp64 = ir.DoubleType()


class HeavyDBBufferType(typesystem.Type):
    """Typesystem type class for HeavyDB buffer structures.
    """
    @property
    def numba_type(self):
        return BufferType

    @property
    def numba_pointer_type(self):
        return BufferPointer

    @classmethod
    def preprocess_args(cls, args):
        assert len(args) == 1, args
        assert len(args[0]) == 1, args
        element_type = args[0][0]
        if not isinstance(element_type, typesystem.Type):
            element_type = typesystem.Type.fromobject(element_type)
        return ((element_type,),)

    @property
    def element_type(self):
        return self[0][0]

    @property
    def buffer_extra_members(self):
        return ()

    @property
    def custom_params(self):
        return {
            'NumbaType': self.numba_type,
            'NumbaPointerType': self.numba_pointer_type,
        }

    def tonumba(self, bool_is_int8=None):
        ptr_t = typesystem.Type(self.element_type, '*', name='ptr')
        size_t = typesystem.Type.fromstring('size_t sz')
        extra_members = tuple(map(typesystem.Type.fromobject, self.buffer_extra_members))
        buffer_type = typesystem.Type(
            ptr_t,
            size_t,
            *extra_members
        )
        buffer_type.params(other=None, **self.custom_params)
        numba_type = buffer_type.tonumba(bool_is_int8=True)
        return self.numba_pointer_type(numba_type)


class BufferType(types.IterableType):
    """Numba type class for HeavyDB buffer structures.
    """

    @property
    def eltype(self):
        """
        Return buffer element dtype.
        """
        return self.members[0].dtype

    @property
    def iterator_type(self):
        return BufferIteratorType(self)


class BufferPointer(types.IterableType):
    """Numba type class for pointers to HeavyDB buffer structures.

    We are not deriving from CPointer because BufferPointer getitem is
    used to access the data stored in Buffer ptr member.
    """
    mutable = True
    return_as_first_argument = True

    def __init__(self, dtype):
        self.dtype = dtype    # struct dtype
        self.eltype = dtype.eltype  # buffer element dtype
        name = "%s[%s]*" % (type(self).__name__, dtype)
        super().__init__(name)

    @property
    def key(self):
        return self.dtype

    @property
    def iterator_type(self):
        return BufferIteratorType(self)

    def deepcopy(self, context, builder, val, retptr):
        raise NotImplementedError


class BufferIteratorType(types.SimpleIteratorType):

    def __init__(self, buffer_type):
        name = f"iter_buffer({buffer_type})"
        self.buffer_type = buffer_type
        super().__init__(name, self.buffer_type.eltype)


@datamodel.register_default(BufferIteratorType)
class BufferPointerIteratorModel(datamodel.StructModel):
    def __init__(self, dmm, fe_type):
        members = [('index', types.EphemeralPointer(types.uintp)),
                   ('buffer', fe_type.buffer_type)]
        super(BufferPointerIteratorModel, self).__init__(dmm, fe_type, members)


class BufferMeta(HeavyDBMetaType):
    pass


class Buffer(object, metaclass=BufferMeta):
    """Represents HeavyDB Buffer that can be constructed within UDF/UDTFs.
    """


@datamodel.register_default(BufferPointer)
class BufferPointerModel(datamodel.models.PointerModel):
    """Base class for HeavyDB buffer pointer models.

    Subclasses should register the model for the corresponding
    BufferPointer subclass, for instance::

      @datamodel.register_default(ArrayPointer)
      class ArrayPointerModel(BufferPointerModel):
          pass
    """


def memalloc(context, builder, ptr_type, element_count, element_size):
    alloc_fn_name = 'allocate_varlen_buffer'
    alloc_fnty = ir.FunctionType(int8_t.as_pointer(), [int64_t, int64_t])
    alloc_fn = cgutils.get_or_insert_function(builder.module,
                                              alloc_fnty,
                                              alloc_fn_name)
    ptr8 = builder.call(alloc_fn, [element_count, element_size])
    ptr = builder.bitcast(ptr8, context.get_value_type(ptr_type))
    return ptr


def heavydb_buffer_constructor(context, builder, sig, args):
    """

    Usage:

      extending.lower_builtin(MyBuffer, numba.types.Integer, ...)(heavydb_buffer_constructor)

    will enable creating MyBuffer instance from a HeavyDB UDF/UDTF definition:

      b = MyBuffer(<size>, ...)

    """

    ptr_type, sz_type = sig.return_type.dtype.members[:2]
    if len(sig.return_type.dtype.members) > 2:
        assert len(sig.return_type.dtype.members) == 3
        null_type = sig.return_type.dtype.members[2]
    else:
        null_type = None
    assert isinstance(args[0].type, ir.IntType), (args[0].type)
    element_count = builder.zext(args[0], int64_t)
    element_size = int64_t(ptr_type.dtype.bitwidth // 8)

    ptr = memalloc(context, builder, ptr_type, element_count, element_size)

    llty = context.get_value_type(sig.return_type.dtype)
    st_size = context.get_abi_sizeof(llty)
    st_ptr = builder.bitcast(allocate_varlen_buffer(builder, int64_t(st_size),
                                                    int64_t(1)),
                             llty.as_pointer())

    zero, one, two = int32_t(0), int32_t(1), int32_t(2)
    builder.store(ptr, builder.gep(st_ptr, [zero, zero]))
    builder.store(element_count, builder.gep(st_ptr, [zero, one]))

    if null_type is not None:
        is_zero = builder.icmp_signed('==', element_count, int64_t(0))
        with builder.if_else(is_zero) as (then, orelse):
            with then:
                is_null = context.get_value_type(null_type)(1)
                builder.store(is_null, builder.gep(st_ptr, [zero, two]))
            with orelse:
                is_null = context.get_value_type(null_type)(0)
                builder.store(is_null, builder.gep(st_ptr, [zero, two]))
    return st_ptr


@extending.intrinsic
def heavydb_buffer_ptr_get_ptr_(typingctx, data):
    eltype = data.eltype
    ptrtype = types.CPointer(eltype)
    sig = ptrtype(data)

    def codegen(context, builder, signature, args):
        data,  = args
        rawptr = cgutils.alloca_once_value(builder, value=data)
        struct = builder.load(builder.gep(rawptr,
                                          [int32_t(0)]))
        return builder.load(builder.gep(struct, [int32_t(0), int32_t(0)]))

    return sig, codegen


@extending.intrinsic
def heavydb_buffer_get_ptr_(typingctx, data):
    eltype = data.eltype
    ptrtype = types.CPointer(eltype)
    sig = ptrtype(data)

    def codegen(context, builder, signature, args):
        data, = args
        assert data.opname == 'load'
        struct = data.operands[0]
        return builder.load(builder.gep(struct, [int32_t(0), int32_t(0)]))

    return sig, codegen


@extending.intrinsic
def heavydb_buffer_ptr_item_get_ptr_(typingctx, data, index):
    eltype = data.eltype
    ptrtype = types.CPointer(eltype)
    sig = ptrtype(data, index)

    def codegen(context, builder, signature, args):
        data, index = args
        rawptr = cgutils.alloca_once_value(builder, value=data)
        struct = builder.load(builder.gep(rawptr, [int32_t(0)]))
        ptr = builder.load(builder.gep(struct, [int32_t(0), int32_t(0)]))
        return builder.gep(ptr, [index])

    return sig, codegen


@extending.overload_method(BufferPointer, 'ptr')
def heavydb_buffer_get_ptr(x, index=None):
    if isinstance(x, BufferPointer):
        if cgutils.is_nonelike(index):
            def impl(x, index=None):
                return heavydb_buffer_ptr_get_ptr_(x)
        else:
            def impl(x, index=None):
                return heavydb_buffer_ptr_item_get_ptr_(x, index)
        return impl
    if isinstance(x, BufferType):
        if cgutils.is_nonelike(index):
            def impl(x, index=None):
                return heavydb_buffer_get_ptr_(x)
        else:
            raise NotImplementedError(f'heavydb_buffer_item_get_ptr_({x}, {index})')
        return impl


@extending.intrinsic
def heavydb_buffer_ptr_len_(typingctx, data):
    sig = types.int64(data)

    def codegen(context, builder, signature, args):
        data, = args
        rawptr = cgutils.alloca_once_value(builder, value=data)
        struct = builder.load(builder.gep(rawptr,
                                          [int32_t(0)]))
        return builder.load(builder.gep(
            struct, [int32_t(0), int32_t(1)]))
    return sig, codegen


@extending.intrinsic
def heavydb_buffer_len_(typingctx, data):
    sig = types.int64(data)

    def codegen(context, builder, signature, args):
        data, = args
        return builder.extract_value(data, [1])

    return sig, codegen


@extending.overload(len)
def heavydb_buffer_len(x):
    if isinstance(x, BufferPointer):
        return lambda x: heavydb_buffer_ptr_len_(x)
    if isinstance(x, BufferType):
        return lambda x: heavydb_buffer_len_(x)


@extending.intrinsic
def heavydb_buffer_ptr_getitem_(typingctx, data, index):
    sig = data.eltype(data, index)

    def codegen(context, builder, signature, args):
        data, index = args
        rawptr = cgutils.alloca_once_value(builder, value=data)
        buf = builder.load(builder.gep(rawptr, [int32_t(0)]))
        ptr = builder.load(builder.gep(
            buf, [int32_t(0), int32_t(0)]))
        res = builder.load(builder.gep(ptr, [index]))

        return res
    return sig, codegen


@extending.intrinsic
def heavydb_buffer_getitem_(typingctx, data, index):
    eltype = data.eltype
    sig = eltype(data, index)

    def codegen(context, builder, signature, args):
        data, index = args
        ptr = builder.extract_value(data, [0])
        res = builder.load(builder.gep(ptr, [index]))

        return res
    return sig, codegen


@extending.overload(operator.getitem)
def heavydb_buffer_getitem(x, i):
    if isinstance(x, BufferPointer):
        return lambda x, i: heavydb_buffer_ptr_getitem_(x, i)
    if isinstance(x, BufferType):
        return lambda x, i: heavydb_buffer_getitem_(x, i)


@extending.intrinsic
def heavydb_buffer_ptr_setitem_(typingctx, data, index, value):
    sig = types.none(data, index, value)

    def codegen(context, builder, sig, args):
        zero = int32_t(0)

        data, index, value = args

        rawptr = cgutils.alloca_once_value(builder, value=data)
        ptr = builder.load(rawptr)

        buf = builder.load(builder.gep(ptr, [zero, zero]))
        # [rbc issue-197] Numba promotes operations like
        # int32(a) + int32(b) to int64
        fromty = sig.args[2]
        toty = sig.args[0].eltype
        value = context.cast(builder, value, fromty, toty)
        builder.store(value, builder.gep(buf, [index]))

    return sig, codegen


@extending.intrinsic
def heavydb_buffer_setitem_(typingctx, data, index, value):
    sig = types.none(data, index, value)

    def codegen(context, builder, sig, args):
        data, index, value = args
        ptr = builder.extract_value(data, [0])

        fromty = sig.args[2]
        toty = sig.args[0].eltype
        value = context.cast(builder, value, fromty, toty)

        builder.store(value, builder.gep(ptr, [index]))

    return sig, codegen


@extending.overload(operator.setitem)
def heavydb_buffer_setitem(a, i, v):
    if isinstance(a, BufferPointer):
        return lambda a, i, v: heavydb_buffer_ptr_setitem_(a, i, v)
    if isinstance(a, BufferType):
        return lambda a, i, v: heavydb_buffer_setitem_(a, i, v)


@extending.intrinsic
def heavydb_buffer_is_null_(typingctx, data):
    sig = types.int8(data)

    def codegen(context, builder, sig, args):
        rawptr = cgutils.alloca_once_value(builder, value=args[0])
        ptr = builder.load(rawptr)
        return builder.load(builder.gep(ptr, [int32_t(0), int32_t(2)]))

    return sig, codegen


@extending.intrinsic
def heavydb_buffer_set_null_(typingctx, data):
    sig = types.none(data)

    def codegen(context, builder, sig, args):
        [ptr] = args
        zero, two = int32_t(0), int32_t(2)
        builder.store(int8_t(1), builder.gep(ptr, [zero, two]))

    return sig, codegen


@extending.intrinsic
def heavydb_buffer_idx_is_null_(typingctx, col_var, row_idx):
    T = col_var.eltype
    sig = types.boolean(col_var, row_idx)

    def codegen(context, builder, signature, args):
        ptr, index = args
        data = builder.extract_value(builder.load(ptr), [0])
        res = builder.load(builder.gep(data, [index]))

        target_info = TargetInfo()
        null_value = target_info.null_values[str(T)]
        # The server sends numbers as unsigned values rather than signed ones.
        # Thus, 129 should be read as -127 (overflow). See rbc issue #254
        nv = ir.Constant(ir.IntType(T.bitwidth), null_value)

        # importing it here in case in the future someone tries to import
        # buffer on timestamp.py
        from .timestamp import TimestampNumbaType
        if isinstance(signature.args[0].eltype, TimestampNumbaType):
            res = builder.extract_value(res, 0)

        if isinstance(T, types.Float):
            res = builder.bitcast(res, nv.type)

        return builder.icmp_signed('==', res, nv)

    def codegen_column_array(context, builder, sig, args):
        ptr, index = args
        i8p = int8_t.as_pointer()
        fnty = ir.FunctionType(int8_t, [i8p, int64_t])
        isNull = cgutils.get_or_insert_function(builder.module, fnty,
                                                "ColumnArray_isNull")
        flatbuffer = builder.extract_value(builder.load(ptr), [0])
        return builder.call(isNull, [flatbuffer, index])

    if isinstance(T, BufferPointer):
        return sig, codegen_column_array
    else:
        return sig, codegen


# "BufferPointer.is_null" checks if a given array or column is null
# as opposed to "BufferType.is_null" that checks if an index in a
# column is null
@extending.overload_method(BufferPointer, 'is_null')
def heavydb_buffer_is_null(x, row_idx=None):
    if isinstance(x, BufferPointer):
        if cgutils.is_nonelike(row_idx):
            def impl(x, row_idx=None):
                return heavydb_buffer_is_null_(x)
        else:
            def impl(x, row_idx=None):
                return heavydb_buffer_idx_is_null_(x, row_idx)
        return impl


def get_null_value(buffer):
    T = buffer.eltype

    target_info = TargetInfo()
    null_value = target_info.null_values[f'{T}']

    # null_values are stored as uin64 values but when assigning to
    # i.e. Column<double>, they null value need to be bitcasted to
    # double.
    toty = T.key
    conv_table = {
        'Timestamp': 'int64',
        'boolean8': 'int8',
        'boolean1': 'int8',
    }
    toty = conv_table.get(toty, toty)

    # The server sends numbers as unsigned values rather than signed ones.
    # Thus, 129 should be read as -127 (overflow). See rbc issue #254
    bitwidth = T.bitwidth
    null_value = np.dtype(f'uint{bitwidth}').type(null_value).view(toty)
    return null_value


@extending.overload_method(BufferPointer, 'set_null')
def heavydb_buffer_set_null(x, row_idx=None):
    if isinstance(x, BufferPointer):
        if cgutils.is_nonelike(row_idx):
            def impl(x, row_idx=None):
                return heavydb_buffer_set_null_(x)
        else:
            null_value = get_null_value(x)

            def impl(x, row_idx=None):
                x[row_idx] = null_value
            return impl
        return impl


@extending.overload(operator.eq)
def dtype_eq(a, b):
    if isinstance(a, types.DTypeSpec) and isinstance(b, types.DTypeSpec):
        eq = (a == b)

        def impl(a, b):
            return eq
        return impl


@extending.overload_attribute(BufferPointer, 'dtype')
def heavydb_buffer_dtype(x):
    if isinstance(x, BufferPointer):
        dtype = x.eltype

        def impl(x):
            return dtype
        return impl


@extending.lower_builtin('iternext', BufferIteratorType)
@imputils.iternext_impl(imputils.RefType.UNTRACKED)
def iternext_BufferPointer(context, builder, sig, args, result):
    [iterbufty] = sig.args
    [bufiter] = args

    iterval = context.make_helper(builder, iterbufty, value=bufiter)

    buf = iterval.buffer
    idx = builder.load(iterval.index)

    if buf.type.is_pointer:
        col = context.make_helper(builder, iterbufty.buffer_type.dtype,
                                  value=builder.load(buf))
    else:
        col = context.make_helper(builder, iterbufty.buffer_type,
                                  value=buf)
    count = col.sz

    is_valid = builder.icmp_signed('<', idx, count)
    result.set_valid(is_valid)

    with builder.if_then(is_valid):
        getitem_fn = context.typing_context.resolve_value_type(operator.getitem)
        getitem_sig = iterbufty.buffer_type.eltype(iterbufty.buffer_type, types.intp)
        getitem_fn.get_call_type(context.typing_context, getitem_sig.args, {})
        getitem_out = context.get_function(getitem_fn, getitem_sig)(builder, [buf, idx])
        result.yield_(getitem_out)
        nidx = builder.add(idx, context.get_constant(types.intp, 1))
        builder.store(nidx, iterval.index)


@extending.lower_builtin('getiter', BufferType)
@extending.lower_builtin('getiter', BufferPointer)
def getiter_buffer_pointer(context, builder, sig, args):
    [buffer] = args

    iterobj = context.make_helper(builder, sig.return_type)

    # set the index to zero
    zero = context.get_constant(types.uintp, 0)
    indexptr = cgutils.alloca_once_value(builder, zero)

    iterobj.index = indexptr

    # wire in the buffer type data
    iterobj.buffer = buffer

    res = iterobj._getvalue()
    return imputils.impl_ret_new_ref(context, builder, sig.return_type, res)



"""
Array API for rbc.
"""

from .constants import *  # noqa: F401, F403
from .creation_functions import *  # noqa: F401, F403
from .datatypes import *  # noqa: F401, F403
from .data_type_functions import *  # noqa: F401, F403
from .elementwise_functions import *  # noqa: F401, F403
from .linear_algebra_functions import *  # noqa: F401, F403
from .manipulation_functions import *  # noqa: F401, F403
from .searching_functions import *  # noqa: F401, F403
from .set_functions import *  # noqa: F401, F403
from .sorting_functions import *  # noqa: F401, F403
from .statistical_functions import *  # noqa: F401, F403
from .utility_functions import *  # noqa: F401, F403

__all__ = [s for s in dir() if not s.startswith("_")]



"""RBC GeoMultiPolygon type that corresponds to HeavyDB type GeoMultiPolygon.
"""

__all__ = ["HeavyDBGeoMultiPolygonType", "GeoMultiPolygon"]

from typing import List

from numba.core import extending
from numba.core import types as nb_types

from . import geopolygon
from .geo_nested_array import (GeoNestedArray, GeoNestedArrayNumbaType,
                               HeavyDBGeoNestedArray,
                               heavydb_geo_fromCoords_vec3,
                               heavydb_geo_toCoords_vec3)


class GeoMultiPolygonNumbaType(GeoNestedArrayNumbaType):
    def __init__(self, name):
        super().__init__(name)


class HeavyDBGeoMultiPolygonType(HeavyDBGeoNestedArray):
    """Typesystem type class for HeavyDB buffer structures."""

    @property
    def numba_type(self):
        return GeoMultiPolygonNumbaType

    @property
    def type_name(self):
        return "GeoMultiPolygon"

    @property
    def item_type(self):
        return "GeoPolygon"


class GeoMultiPolygon(GeoNestedArray):
    """
    RBC ``GeoMultiPolygon`` type that corresponds to HeavyDB type
    ``MULTIPOLYGON``.

    .. code-block:: c

        struct MultiPolygon {
            int8_t* flatbuffer_;
            int64_t index_[4];
            int64_t n_;
        }
    """

    def __getitem__(self, index: int) -> 'geopolygon.GeoPolygon':
        """
        Return the ``POLYGON`` at given index
        """

    def get_item(self, index: int) -> 'geopolygon.GeoPolygon':
        """
        Return the ``POLYGON`` at given index
        """

    def to_coords(self) -> List[List[List[float]]]:
        """
        .. note::
            Only available on ``CPU``
        """

    def from_coords(self, coords: List[List[List[float]]]) -> None:
        """
        .. note::
            Only available on ``CPU``
        """


@extending.overload_method(GeoMultiPolygonNumbaType, "from_coords")
def heavydb_geomultipolygon_fromCoords(geo, lst):
    if isinstance(geo, GeoMultiPolygonNumbaType) and isinstance(lst, nb_types.List):
        return heavydb_geo_fromCoords_vec3(geo, lst)


@extending.overload_method(GeoMultiPolygonNumbaType, "to_coords")
def heavydb_geolinestring_toCoords(geo):
    return heavydb_geo_toCoords_vec3(geo)



"""Implement Heavydb Column<GeoMultiLineString> type support

Heavydb Column<GeoMultiLineString> type is the type of input/output column arguments in
UDTFs.
"""

__all__ = [
    "HeavyDBOutputColumnGeoMultiLineStringType",
    "HeavyDBColumnGeoMultiLineStringType",
    "ColumnGeoMultiLineString",
]

from . import geomultilinestring
from .column_flatbuffer import (ColumnFlatBuffer, HeavyDBColumnFlatBufferType,
                                HeavyDBOutputColumnFlatBufferType)


class ColumnGeoMultiLineString(ColumnFlatBuffer):
    """
    RBC ``Column<GeoMultiLineString>`` type that corresponds to HeavyDB
    ``COLUMN<MULTILINESTRING>``

    In HeavyDB, a Column of type ``GeoMultiLineString`` is represented as follows:

    .. code-block:: c

        {
            int8_t* flatbuffer;
            int64_t size;
        }

    """

    def __getitem__(self, index: int) -> 'geomultilinestring.GeoMultiLineString':
        """
        Return ``self[index]``

        .. note::
            Only available on ``CPU``
        """

    def get_item(self, index: int) -> 'geomultilinestring.GeoMultiLineString':
        """
        Return ``self[index]``

        .. note::
            Only available on ``CPU``
        """

    def set_item(self, index: int, buf: 'geomultilinestring.GeoMultiLineString') -> None:
        """
        Set line from a buffer of point coordindates

        .. note::
            Only available on ``CPU``
        """

    def __setitem__(self, index: int, buf: 'geomultilinestring.GeoMultiLineString') -> None:
        """
        Set line from a buffer of point coordindates

        .. note::
            Only available on ``CPU``
        """


class HeavyDBColumnGeoMultiLineStringType(HeavyDBColumnFlatBufferType):
    """ """

    @property
    def type_name(self):
        return "GeoMultiLineString"


class HeavyDBOutputColumnGeoMultiLineStringType(
    HeavyDBColumnGeoMultiLineStringType, HeavyDBOutputColumnFlatBufferType
):
    """Heavydb OutputColumn type for RBC typesystem.

    OutputColumn<GeoMultiLineString> is the same as Column<GeoMultiLineString> but introduced to
    distinguish the input and output arguments of UDTFs.
    """



import pytest
import numpy as np
from rbc.heavydb import Array
from rbc.tests import heavydb_fixture
from numba import types as nb_types
from rbc.stdlib import array_api  # noqa: F401
import operator


rbc_heavydb = pytest.importorskip('rbc.heavydb')
available_version, reason = rbc_heavydb.is_available()
pytestmark = pytest.mark.skipif(not available_version, reason=reason)


@pytest.fixture(scope='module')
def heavydb():
    for o in heavydb_fixture(globals(), load_test_data=False):
        define(o)
        yield o


operator_methods = [
    ('abs', (6,), np.arange(6)),
    ('add', (6,), np.full(6, 5)),
    ('and_bw', (6,), [0, 0, 2, 2, 0, 0]),
    ('countOf', (6, 3, 4), 0),
    ('countOf', (6, 3, 3), 6),
    ('eq', (6, 3), [0, 0, 0, 1, 0, 0]),
    ('eq_array', (6, 3), True),
    ('floordiv', (6,), [3, 2, 2, 2, 2, 1]),
    ('floordiv2', (6,), [3.0, 2.0, 2.0, 2.0, 2.0, 1.0]),
    ('ge', (6, 3), [0, 0, 0, 1, 1, 1]),
    ('ge_array', (6, 3), True),
    ('gt', (6, 3), [0, 0, 0, 0, 1, 1]),
    ('gt_array', (6, 3), False),
    ('iadd', (6,), [1, 2, 3, 4, 5, 6]),
    ('iand', (6,), [0, 0, 2, 2, 0, 0]),
    ('ifloordiv', (6,), [3, 2, 2, 2, 2, 1]),
    ('ifloordiv2', (6,), [3, 2, 2, 2, 2, 1]),
    ('ilshift', (6,), [0, 16, 16, 12, 8, 5]),
    ('imul', (6,), [0, 4, 6, 6, 4, 0]),
    ('ior', (6,), [5, 5, 3, 3, 5, 5]),
    ('isub', (6,), [-5, -3, -1, 1, 3, 5]),
    ('ipow', (6,), [1, 32, 81, 64, 25, 6]),
    ('irshift', (6,), [0, 0, 0, 0, 2, 5]),
    ('itruediv', (6,), [3, 2, 2, 2, 2, 1]),
    ('itruediv2', (6,), [3.3333333333333335, 2.75, 2.4, 2.1666666666666665, 2.0, 1.875]),  # noqa: E501
    ('imod', (6,), [0, 4, 1, 5, 2, 6]),
    ('ixor', (6,), [5, 5, 1, 1, 5, 5]),
    ('in', (6, 3), True),
    ('is', (6, 3), True),
    ('is_not', (6, 3), False),
    ('is_not2', (6, 3), True),
    ('le', (6, 3), [1, 1, 1, 1, 0, 0]),
    ('le_array', (6, 3), True),
    ('lshift', (6,), [0, 16, 16, 12, 8, 5]),
    ('lt', (6, 3), [1, 1, 1, 0, 0, 0]),
    ('lt_array', (6, 3), False),
    ('mul', (6,), [0, 4, 6, 6, 4, 0]),
    ('mod', (6,), [0, 4, 1, 5, 2, 6]),
    ('ne', (6, 3), [1, 1, 1, 0, 1, 1]),
    ('ne_array', (6, 3), False),
    ('neg', (6,), [0, -1, -2, -3, -4, -5]),
    ('not_in', (6, 3), False),
    ('or_bw', (6,), [5, 5, 3, 3, 5, 5]),
    ('pos', (6,), [0, -1, -2, -3, -4, -5]),
    ('pow', (6,), [1, 32, 81, 64, 25, 6]),
    ('rshift', (6,), [0, 0, 0, 0, 2, 5]),
    ('sub', (6,), [-5, -3, -1, 1, 3, 5]),
    ('truediv', (6,), [3, 2, 2, 2, 2, 1]),
    ('truediv2', (6,), [3.3333333333333335, 2.75, 2.4, 2.1666666666666665, 2.0, 1.875]),  # noqa: E501
    ('xor', (6,), [5, 5, 1, 1, 5, 5]),
]


def define(heavydb):

    @heavydb('int32[](int64)')
    def operator_abs(size):
        a = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(-i)
        return abs(a)

    @heavydb('int32[](int64)')
    def operator_add(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
            b[i] = nb_types.int32(size-i-1)
        return operator.add(a, b)

    @heavydb('int32[](int64)')
    def operator_and_bw(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
            b[i] = nb_types.int32(size-i-1)
        return operator.and_(a, b)

    @heavydb('int64(int64, int64, int64)', devices=['CPU'])
    def operator_countOf(size, fill_value, b):
        a = Array(size, 'int64')
        for i in range(size):
            a[i] = fill_value
        return operator.countOf(a, b)

    @heavydb('int8[](int64, int32)')
    def operator_eq(size, v):
        a = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
        return a == v

    @heavydb('bool(int64, int32)', devices=['CPU'])
    def operator_eq_array(size, v):
        a = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
        return a == a

    @heavydb('int32[](int64)')
    def operator_floordiv(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i+10)
            b[i] = nb_types.int32(i+3)
        return operator.floordiv(a, b)

    @heavydb('double[](int64)')
    def operator_floordiv2(size):
        a = Array(size, 'double')
        b = Array(size, 'double')
        for i in range(size):
            a[i] = nb_types.double(i+10)
            b[i] = nb_types.double(i+3)
        return operator.floordiv(a, b)

    @heavydb('int8[](int64, int32)')
    def operator_ge(size, v):
        a = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
        return a >= v

    @heavydb('bool(int64, int32)', devices=['CPU'])
    def operator_ge_array(size, v):
        a = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
        return a >= a

    @heavydb('int8[](int64, int32)')
    def operator_gt(size, v):
        a = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
        return a > v

    @heavydb('bool(int64, int32)', devices=['CPU'])
    def operator_gt_array(size, v):
        a = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
        return a > a

    @heavydb('int32[](int64)')
    def operator_iadd(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
            b[i] = nb_types.int32(1)
        operator.iadd(a, b)
        return a

    @heavydb('int32[](int64)')
    def operator_iand(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
            b[i] = nb_types.int32(size-i-1)
        operator.iand(a, b)
        return a

    @heavydb('int32[](int64)')
    def operator_ifloordiv(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i+10)
            b[i] = nb_types.int32(i+3)
        operator.ifloordiv(a, b)
        return a

    @heavydb('double[](int64)')
    def operator_ifloordiv2(size):
        a = Array(size, 'double')
        b = Array(size, 'double')
        for i in range(size):
            a[i] = nb_types.double(i+10)
            b[i] = nb_types.double(i+3)
        operator.ifloordiv(a, b)
        return a

    @heavydb('int32[](int64)')
    def operator_ilshift(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
            b[i] = nb_types.int32(size-i-1)
        operator.ilshift(a, b)
        return a

    @heavydb('int32[](int64)')
    def operator_imul(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
            b[i] = nb_types.int32(size-i-1)
        operator.imul(a, b)
        return a

    @heavydb('int32[](int64)')
    def operator_ior(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
            b[i] = nb_types.int32(size-i-1)
        operator.ior(a, b)
        return a

    @heavydb('int32[](int64)')
    def operator_isub(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
            b[i] = nb_types.int32(size-i-1)
        operator.isub(a, b)
        return a

    @heavydb('int32[](int64)')
    def operator_ipow(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i+1)
            b[i] = nb_types.int32(size-i)
        operator.ipow(a, b)
        return a

    @heavydb('int32[](int64)')
    def operator_irshift(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
            b[i] = nb_types.int32(size-i-1)
        operator.irshift(a, b)
        return a

    @heavydb('int32[](int64)')
    def operator_itruediv(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i+10)
            b[i] = nb_types.int32(i+3)
        operator.itruediv(a, b)
        return a

    @heavydb('double[](int64)')
    def operator_itruediv2(size):
        a = Array(size, 'double')
        b = Array(size, 'double')
        for i in range(size):
            a[i] = nb_types.double(i+10)
            b[i] = nb_types.double(i+3)
        operator.itruediv(a, b)
        return a

    @heavydb('int32[](int64)')
    def operator_imod(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i * 123)
            b[i] = nb_types.int32(7)
        operator.imod(a, b)
        return a

    @heavydb('int32[](int64)')
    def operator_ixor(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
            b[i] = nb_types.int32(size-i-1)
        operator.ixor(a, b)
        return a

    @heavydb('int8(int64, int32)', devices=['CPU'])
    def operator_in(size, v):
        a = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
        return v in a

    @heavydb('int8(int64, int32)', devices=['CPU'])
    def operator_is(size, v):
        a = Array(size, 'int32')
        a.fill(v)
        return a is a

    @heavydb('int8(int64, int32)', devices=['CPU'])
    def operator_is_not(size, v):
        a = Array(size, 'int32')
        a.fill(v)
        return a is not a

    @heavydb('int8(int64, int32)', devices=['CPU'])
    def operator_is_not2(size, v):
        a = Array(size, 'int32')
        a.fill(v)
        b = Array(size, 'int32')
        b.fill(v)
        return a is not b

    @heavydb('int8[](int64, int32)')
    def operator_le(size, v):
        a = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
        return a <= v

    @heavydb('bool(int64, int32)', devices=['CPU'])
    def operator_le_array(size, v):
        a = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
        return a <= a

    @heavydb('int32[](int64)')
    def operator_lshift(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
            b[i] = nb_types.int32(size-i-1)
        return operator.lshift(a, b)

    @heavydb('int8[](int64, int32)')
    def operator_lt(size, v):
        a = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
        return a < v

    @heavydb('bool(int64, int32)', devices=['CPU'])
    def operator_lt_array(size, v):
        a = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
        return a < a

    @heavydb('int32[](int64)')
    def operator_mul(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
            b[i] = nb_types.int32(size-i-1)
        return operator.mul(a, b)

    @heavydb('int32[](int64)')
    def operator_mod(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i * 123)
            b[i] = nb_types.int32(7)
        return operator.mod(a, b)

    @heavydb('int8[](int64, int32)')
    def operator_ne(size, v):
        a = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
        return a != v

    @heavydb('bool(int64, int32)', devices=['CPU'])
    def operator_ne_array(size, v):
        a = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
        return a != a

    @heavydb('int32[](int64)')
    def operator_neg(size):
        a = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
        return operator.neg(a)

    @heavydb('int8(int64, int32)', devices=['CPU'])
    def operator_not_in(size, v):
        a = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
        return v not in a

    @heavydb('int32[](int64)')
    def operator_or_bw(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
            b[i] = nb_types.int32(size-i-1)
        return operator.or_(a, b)

    @heavydb('int32[](int64)')
    def operator_pos(size):
        a = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(-i)
        return operator.pos(a)

    @heavydb('int32[](int64)')
    def operator_pow(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i+1)
            b[i] = nb_types.int32(size-i)
        return operator.pow(a, b)

    @heavydb('int32[](int64)')
    def operator_rshift(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
            b[i] = nb_types.int32(size-i-1)
        return operator.rshift(a, b)

    @heavydb('int32[](int64)')
    def operator_sub(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
            b[i] = nb_types.int32(size-i-1)
        return operator.sub(a, b)

    @heavydb('int32[](int64)')
    def operator_truediv(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i+10)
            b[i] = nb_types.int32(i+3)
        return operator.truediv(a, b)

    @heavydb('double[](int64)')
    def operator_truediv2(size):
        a = Array(size, 'double')
        b = Array(size, 'double')
        for i in range(size):
            a[i] = nb_types.double(i+10)
            b[i] = nb_types.double(i+3)
        return operator.truediv(a, b)

    @heavydb('int32[](int64)')
    def operator_xor(size):
        a = Array(size, 'int32')
        b = Array(size, 'int32')
        for i in range(size):
            a[i] = nb_types.int32(i)
            b[i] = nb_types.int32(size-i-1)
        return operator.xor(a, b)


@pytest.mark.parametrize("suffix, args, expected", operator_methods,
                         ids=[item[0] for item in operator_methods])
def test_array_operators(heavydb, suffix, args, expected):
    query = 'select operator_{suffix}'.format(**locals()) + \
            '(' + ', '.join(map(str, args)) + ')'
    _, result = heavydb.sql_execute(query)
    out = list(result)[0]

    if suffix in ['in', 'not_in']:
        assert (expected == out[0]), 'operator_' + suffix
    elif '_array' in suffix:
        assert (expected == out[0]), 'operator_' + suffix
    else:
        assert np.array_equal(expected, out[0]), 'operator_' + suffix



"""
Array API specification for data types.

https://data-apis.org/array-api/latest/API_specification/data_types.html
"""

__all__ = [
    'Array',
    'bool',
    'int8',
    'int16',
    'int32',
    'int64',
    'uint8',
    'uint16',
    'uint32',
    'uint64',
    'float32',
    'float64'
]

# NOTE: currently the code lives in rbc.heavydb, but eventually we
# should move it here and leave rbc.heavydb.Array only for backwards
# compatibility
from rbc.heavydb import Array

# array API data types
from numba.types import (
    boolean as bool,
    int8,
    int16,
    int32,
    int64,
    uint8,
    uint16,
    uint32,
    uint64,
    float32,
    float64)



import os
import pytest
from rbc.tests import heavydb_fixture


@pytest.fixture(scope="module")
def heavydb():
    for o in heavydb_fixture(globals(), minimal_version=(5, 7), suffices=['']):
        o.base_name = os.path.splitext(os.path.basename(__file__))[0]
        yield o


@pytest.fixture(scope="function")
def create_columns(heavydb):
    heavydb.require_version((5, 7), "Requires heavydb-internal PR 5492")

    for size in (8, 16, 32):
        table_name = f"{heavydb.base_name}_{size}"
        base = f"base_{size}"
        other = f"other_{size}"
        derived = f"derived_{size}"

        heavydb.sql_execute(f"DROP TABLE IF EXISTS {table_name};")

        heavydb.sql_execute(
            f"""
            CREATE TABLE IF NOT EXISTS {table_name} (
                {base} TEXT ENCODING DICT({size}),
                {other} TEXT ENCODING DICT({size}),
                {derived} TEXT,
                SHARED DICTIONARY ({derived}) REFERENCES {table_name}({base})
            );"""
        )

        data = {
            base: ["hello", "foo", "foofoo", "world", "bar", "foo", "foofoo"],
            other: ["a1", "b2", "c3", "d4", "e5", "f6", "g7"],
            derived: ["world", "bar", "hello", "foo", "baz", "hello", "foo"],
        }

        heavydb.load_table_columnar(table_name, **data)

    yield heavydb

    for size in (8, 16, 32):
        table_name = f"{heavydb.base_name}_{size}"
        heavydb.sql_execute(f"DROP TABLE IF EXISTS {table_name}")


@pytest.mark.usefixtures("create_columns")
@pytest.mark.parametrize("size", (32,))
def test_template_text(heavydb, size):

    if heavydb.has_cuda:
        heavydb.require_version(
            (5, 8), "Requires heavydb-internal PR 5809")

    fn = "ct_binding_template"
    table = f"{heavydb.base_name}_{size}"
    base = f"base_{size}"

    query = f"SELECT * FROM table({fn}(cursor(SELECT {base} FROM {table})));"
    _, result = heavydb.sql_execute(query)
    assert list(result) == [
        ("hello",),
        ("foo",),
        ("foofoo",),
        ("world",),
        ("bar",),
        ("foo",),
        ("foofoo",),
    ]


@pytest.mark.parametrize("col", ("i4", "f4"))
def test_template_number(heavydb, col):

    if heavydb.has_cuda:
        heavydb.require_version(
            (5, 8), "Requires heavydb-internal PR 5809")

    fn = "ct_binding_template"
    table = heavydb.table_name

    query = f"SELECT * FROM table({fn}(cursor(SELECT {col} FROM {table})));"
    _, result = heavydb.sql_execute(query)
    assert list(result) == [(0,), (1,), (2,), (3,), (4,)]


@pytest.mark.usefixtures("create_columns")
@pytest.mark.parametrize("size", (8, 16, 32,))
@pytest.mark.parametrize("out", (3,))
def test_template_columnlist_text(heavydb, size, out):

    fn = "ct_binding_columnlist"
    table = f"{heavydb.base_name}_{size}"
    base = f"base_{size}"

    query = f"SELECT * FROM table({fn}(cursor(SELECT {base}, {base}, {base} FROM {table})));"
    _, result = heavydb.sql_execute(query)
    assert list(result) == [(out,)]


@pytest.mark.parametrize("col,out", zip(("i4", "f4", "i2"), (1, 2, 4)))
def test_template_columnlist_number(heavydb, col, out):

    fn = "ct_binding_columnlist"
    table = heavydb.table_name

    query = f"SELECT * FROM table({fn}(cursor(SELECT {col}, {col}, {col} FROM {table})));"
    _, result = heavydb.sql_execute(query)
    assert list(result) == [(out,)]


@pytest.mark.parametrize("col,out", zip(("i4", "f4"), (10, 20)))
def test_template_column_number(heavydb, col, out):
    heavydb.require_version((5, 8), "Requires heavydb-internal PR #5770")

    fn = "ct_binding_column"
    table = heavydb.table_name

    query = f"SELECT * FROM table({fn}(cursor(SELECT {col} FROM {table})));"
    _, result = heavydb.sql_execute(query)
    assert list(result) == [(out,)]



import inspect
import os
import re
import functools
from . import types

include_match = re.compile(r'^\s*include\s+"(?P<filename>[\w._/\\]+)"',
                           flags=re.M).match
namespace_match = re.compile(
    r'^\s*namespace\s+(?P<scope>\w+)\s+(?P<identifier>[\w._]+)',
    flags=re.M).match


def resolve_includes(thrift_content, include_dirs,
                     _namespace_list=None,
                     _included=None):
    if _namespace_list is None:
        recursive = False
        _namespace_list = []
        _included = []
    else:
        recursive = True
    lines = []
    for line in thrift_content.splitlines():
        m = include_match(line)
        if m is not None:
            filename = m.group('filename')
            for p in include_dirs:
                fn = os.path.join(p, filename)
                if os.path.isfile(fn):
                    # print('including {}'.format(fn))
                    afn = os.path.abspath(fn)
                    if afn in _included:
                        break
                    _included.append(afn)
                    content = open(fn).read()
                    d = os.path.abspath(os.path.dirname(fn))
                    content = resolve_includes(content, [d] + include_dirs,
                                               _namespace_list,
                                               _included)
                    lines.append(content)
                    break
            else:
                raise FileNotFoundError('could not find {} in {}'.format(
                    filename, os.pathsep.join(include_dirs)))
            continue
        m = namespace_match(line)
        if m is not None:
            scope_and_identifier = m.group('scope'), m.group('identifier')
            if scope_and_identifier not in _namespace_list:
                _namespace_list.append(scope_and_identifier)
            continue
        lines.append(line)
    if not recursive:
        for scope_and_identifier in _namespace_list:
            lines.insert(0, 'namespace %s %s' % scope_and_identifier)
    content = '\n'.join(lines)
    return content


def dispatchermethod(mth):
    """Decorator for dispatcher methods.

    Method annotations are used to transform the arguments and the
    return values of dispatcher methods to required types.
    """
    signature = inspect.signature(mth)

    @functools.wraps(mth)
    def wrapper(*args):
        thrift = args[0].thrift
        new_args = []
        for arg, (argname, sig) in zip(args, signature.parameters.items()):
            cls = sig.annotation
            if cls == sig.empty:
                if isinstance(arg, thrift.Data):
                    cls = None
                else:
                    cls = type(arg)
            arg = types.toobject(thrift, arg, cls=cls)
            new_args.append(arg)
        r = mth(*new_args)
        tcls = signature.return_annotation
        if tcls == sig.empty:
            if isinstance(r, (int, float, list, set, dict, str, bytes)):
                # types supported by thriftpy2
                tcls = type(r)
            else:
                tcls = thrift.Data
        return types.fromobject(thrift, tcls, r)
    wrapper.signature = signature
    return wrapper



from .array import *  # noqa: F401, F403
from .allocator import *  # noqa: F401, F403
from .column import *  # noqa: F401, F403
from .column_array import *  # noqa: F401, F403
from .buffer import *  # noqa: F401, F403
from .metatype import *  # noqa: F401, F403
from .pipeline import *  # noqa: F401, F403
from .column_list import *  # noqa: F401, F403
from .column_list_array import *  # noqa: F401, F403
from .column_geopoint import *  # noqa: F401, F403
from .column_geomultipoint import *  # noqa: F401, F403
from .column_geolinestring import *  # noqa: F401, F403
from .column_geomultilinestring import *  # noqa: F401, F403
from .column_geopolygon import *  # noqa: F401, F403
from .column_geomultipolygon import *  # noqa: F401, F403
from .table_function_manager import *  # noqa: F401, F403
from .row_function_manager import *  # noqa: F401, F403
from .text_encoding_dict import *  # noqa: F401, F403
from .string_dict_proxy import *  # noqa: F401, F403
from .text_encoding_none import *  # noqa: F401, F403
from .timestamp import *  # noqa: F401, F403
from .day_time_interval import *  # noqa: F401, F403
from .year_month_time_interval import *  # noqa: F401, F403
from .geopoint import *  # noqa: F401, F403
from .geolinestring import *  # noqa: F401, F403
from .geomultilinestring import *  # noqa: F401, F403
from .geomultipoint import *  # noqa: F401, F403
from .geopolygon import *  # noqa: F401, F403
from .geomultipolygon import *  # noqa: F401, F403
from .point2d import *  # noqa: F401, F403
from .remoteheavydb import *  # noqa: F401, F403

from . import mathimpl as math  # noqa: F401
from . import npyimpl as np  # noqa: F401
from . import python_operators as operators  # noqa: F401

__all__ = [s for s in dir() if not s.startswith('_')]



import operator

from rbc.errors import NumbaTypeError
from numba.core import ir, types
from numba.core.compiler import CompilerBase, DefaultPassBuilder
from numba.core.compiler_machinery import FunctionPass, register_pass
from numba.core.untyped_passes import (IRProcessing,
                                       RewriteSemanticConstants,
                                       ReconstructSSA,
                                       DeadBranchPrune,)
from numba.core.typed_passes import PartialTypeInference, DeadCodeElimination


@register_pass(mutates_CFG=False, analysis_only=False)
class CheckRaiseStmts(FunctionPass):
    _name = "check_raise_stmts"

    def __init__(self):
        FunctionPass.__init__(self)

    def run_pass(self, state):
        func_ir = state.func_ir
        for blk in func_ir.blocks.values():
            for _raise in blk.find_insts(ir.Raise):
                msg = ('raise statement is not supported in '
                       'UDF/UDTFs. Please, use `return table_function_error(msg)` '
                       'to raise an error.')
                loc = _raise.loc
                raise NumbaTypeError(msg, loc=loc)
        return False


@register_pass(mutates_CFG=False, analysis_only=False)
class DTypeComparison(FunctionPass):
    _name = "DTypeComparison"

    def __init__(self):
        FunctionPass.__init__(self)

    def is_dtype_comparison(self, func_ir, binop):
        """ Return True if binop is a dtype comparison
        """
        def is_getattr(expr):
            return isinstance(expr, ir.Expr) and expr.op == 'getattr'

        if binop.fn != operator.eq:
            return False

        lhs = func_ir.get_definition(binop.lhs.name)
        rhs = func_ir.get_definition(binop.lhs.name)

        return (is_getattr(lhs) and lhs.attr == 'dtype') or \
               (is_getattr(rhs) and rhs.attr == 'dtype')

    def run_pass(self, state):
        # run as subpipeline
        from numba.core.compiler_machinery import PassManager
        pm = PassManager("subpipeline")
        pm.add_pass(PartialTypeInference, "performs partial type inference")
        pm.finalize()
        pm.run(state)

        mutated = False

        func_ir = state.func_ir
        for block in func_ir.blocks.values():
            for assign in block.find_insts(ir.Assign):
                binop = assign.value
                if not (isinstance(binop, ir.Expr) and binop.op == 'binop'):
                    continue
                if self.is_dtype_comparison(func_ir, binop):
                    var = func_ir.get_assignee(binop)
                    typ = state.typemap.get(var.name, None)
                    if isinstance(typ, types.BooleanLiteral):
                        loc = binop.loc
                        rhs = ir.Const(typ.literal_value, loc)
                        new_assign = ir.Assign(rhs, var, loc)

                        # replace instruction
                        block.insert_after(new_assign, assign)
                        block.remove(assign)
                        mutated = True

        if mutated:
            pm = PassManager("subpipeline")
            # rewrite consts / dead branch pruning
            pm.add_pass(DeadCodeElimination, "dead code elimination")
            pm.add_pass(RewriteSemanticConstants, "rewrite semantic constants")
            pm.add_pass(DeadBranchPrune, "dead branch pruning")
            pm.finalize()
            pm.run(state)
        return mutated


class HeavyDBCompilerPipeline(CompilerBase):
    def define_pipelines(self):
        # define a new set of pipelines (just one in this case) and for ease
        # base it on an existing pipeline from the DefaultPassBuilder,
        # namely the "nopython" pipeline
        pm = DefaultPassBuilder.define_nopython_pipeline(self.state)
        # Add the new pass to run after IRProcessing
        pm.add_pass_after(CheckRaiseStmts, IRProcessing)
        pm.add_pass_after(DTypeComparison, ReconstructSSA)
        # finalize
        pm.finalize()
        # return as an iterable, any number of pipelines may be defined!
        return [pm]



"""https://docs.nvidia.com/cuda/libdevice-users-guide/index.html
"""

from . import make_intrinsic
from numba.cuda import libdevicefuncs  # noqa: E402

for fname, (retty, args) in libdevicefuncs.functions.items():
    argnames = [arg.name for arg in args]
    doc = f"libdevice function {fname}"
    fn = make_intrinsic(fname, retty, argnames, __name__, globals(), doc)



"""
Base class for Column<Geo*> types
"""

__all__ = [
    "HeavyDBOutputColumnFlatBufferType",
    "HeavyDBColumnFlatBufferType",
    "ColumnFlatBuffer",
    "ColumnFlatBufferPointer",
    "ColumnFlatBufferType",
]

import operator
from typing import TypeVar

from llvmlite import ir
from numba.core import cgutils, extending
from numba.core import types as nb_types

from rbc import typesystem
from rbc.external import external

from .column import HeavyDBColumnType, HeavyDBOutputColumnType
from .metatype import HeavyDBMetaType
from .utils import as_voidptr, deref, get_alloca

i1 = ir.IntType(1)
i8 = ir.IntType(8)
i8p = i8.as_pointer()
i8pp = i8p.as_pointer()
i32 = ir.IntType(32)
i64 = ir.IntType(64)
i64p = i64.as_pointer()
void = ir.VoidType()

T = TypeVar("T")


class ColumnFlatBuffer(metaclass=HeavyDBMetaType):
    """
    .. code-block:: c

        {
            int8_t* flatbuffer;
            int64_t size;
        }

    """

    def is_null(self, index: int) -> bool:
        """
        Check if element at ``index`` is null.

        .. note::
            Only available on ``CPU``
        """

    def set_null(self, index: int) -> None:
        """
        Set element at ``index`` to null.

        .. note::
            Only available on ``CPU``
        """

    def get_n_of_values(self) -> int:
        """
        Return the total number of points in a Column<T>

        .. note::
            Only available on ``CPU``
        """

    def size(self) -> int:
        """
        Return the length of the Column<T>

        ..  note::
             Only available on ``CPU``
        """

    def __len__(self) -> int:
        """
        Return the length of the Column<T>

        ..  note::
             Only available on ``CPU``
        """


class HeavyDBColumnFlatBufferType(HeavyDBColumnType):
    def postprocess_type(self):
        return self.params(shorttypename="Column")

    @property
    def numba_type(self):
        return ColumnFlatBufferType

    @property
    def numba_pointer_type(self):
        return ColumnFlatBufferPointer

    @property
    def type_name(self):
        raise NotImplementedError()

    @property
    def element_type(self):
        return typesystem.Type.fromstring("int8_t")

    @property
    def custom_params(self):
        return {
            **super().custom_params,
            "name": f"Column<{self.type_name}>",
        }


class HeavyDBOutputColumnFlatBufferType(
    HeavyDBColumnFlatBufferType, HeavyDBOutputColumnType
):
    """ """


class ColumnFlatBufferType(nb_types.Type):
    def __init__(self, name):
        self.dtype = name.split('<', 1)[1][:-1]
        super().__init__(name)

    @property
    def eltype(self):
        """
        Return buffer element dtype.
        """
        return self.dtype


class ColumnFlatBufferPointer(nb_types.Type):
    """Numba type class for pointers to HeavyDB buffer structures."""

    def __init__(self, dtype):
        self.dtype = dtype  # struct dtype
        self.eltype = dtype.eltype  # buffer element dtype
        super().__init__(f"Column<{dtype.eltype}>*")


@extending.intrinsic
def heavydb_column_len_(typingctx, col):
    sig = nb_types.int64(col)

    def codegen(context, builder, sig, args):
        [col] = args
        return builder.extract_value(col, [1])

    return sig, codegen


@extending.intrinsic
def heavydb_column_getitem_(typingctx, col, index):
    eltype = col.eltype
    LS = typesystem.Type.fromstring(eltype).tonumba()
    sig = LS(col, index)

    def codegen(context, builder, sig, args):
        col, index = args
        fnty = ir.FunctionType(void, [i8p, i64, i8p])
        getItem = cgutils.get_or_insert_function(
            builder.module, fnty, f"Column{eltype}_getItem"
        )
        col_ptr = builder.bitcast(col, i8p)

        # Alloca Geo<T>
        fa = context.make_helper(builder, sig.return_type)
        fa.n_ = fa.n_.type(1)

        # Geo<T> -> void*
        result_ptr = builder.bitcast(fa._getpointer(), i8p)

        # call func
        builder.call(getItem, [col_ptr, index, result_ptr])

        # convert void* -> Geo<T>
        point_type = fa._get_be_type(fa._datamodel)
        return builder.load(builder.bitcast(result_ptr, point_type.as_pointer()))

    return sig, codegen


@extending.overload_method(ColumnFlatBufferPointer, "get_n_of_values")
def heavydb_column_getnofvalues(col):
    getNofValues = external(f"int64_t Column{col.eltype}_getNofValues(int8_t*)")

    if isinstance(col, ColumnFlatBufferPointer):

        def impl(col):
            return getNofValues(as_voidptr(col))

        return impl


@extending.overload(operator.getitem)
@extending.overload_method(ColumnFlatBufferPointer, "get_item")
def heavydb_column_getitem(col, index):
    supported_geotypes = {"GeoLineString", "GeoMultiLineString", "GeoPolygon",
                          "GeoMultiPolygon", "GeoMultiPoint"}
    if isinstance(col, ColumnFlatBufferPointer) and col.eltype in supported_geotypes:

        def impl(col, index):
            return heavydb_column_getitem_(col, index)

        return impl


@extending.overload_method(ColumnFlatBufferPointer, "is_null")
def heavydb_column_ptr_is_null(col, index):
    isNull = external(f"bool Column{col.eltype}_isNull(int8_t*, int64_t)")

    if isinstance(col, ColumnFlatBufferPointer):

        def impl(col, index):
            return isNull(as_voidptr(col), index)

        return impl


@extending.overload_method(ColumnFlatBufferPointer, "set_null")
def heavydb_column_set_null(col, index):
    setNull = external(f"bool Column{col.eltype}_setNull(int8_t*, int64_t)")
    if isinstance(col, ColumnFlatBufferPointer):

        def impl(col, index):
            setNull(as_voidptr(col), index)

        return impl


@extending.overload(operator.setitem)
@extending.overload_method(ColumnFlatBufferPointer, "set_item")
def heavydb_column_set_item(col, index, rhs):
    if isinstance(col, ColumnFlatBufferPointer):
        sig = f"void Column{col.eltype}_setItem(int8_t*, int64_t, int8_t* rhs)"
        setItem = external(sig)

        def impl(col, index, rhs):
            setItem(as_voidptr(col), index, as_voidptr(get_alloca(rhs)))

        return impl


@extending.overload(len)
def heavydb_column_len(col):
    if isinstance(col, ColumnFlatBufferPointer):

        def impl(col):
            return heavydb_column_len_(deref(col))

        return impl



import pytest
from rbc.tests import heavydb_fixture
from rbc.externals.heavydb import set_output_row_size


scalar_types = ['float64', 'int64', 'float32', 'int32', 'int16', 'int8']
table_columns_map = dict(int64='i8', int32='i4', int16='i2', int8='i1',
                         float64='f8', float32='f4')


@pytest.fixture(scope='module')
def heavydb():

    for o in heavydb_fixture(globals(), minimal_version=(5, 7, 1), load_test_data=['']):
        define(o)
        yield o


def define(heavydb):

    @heavydb('int32(Column<T>, OutputColumn<T>)', T=scalar_types, devices=['cpu'])
    def column_set_output_row_size(inp, out):
        sz = len(inp)
        set_output_row_size(sz)
        for i in range(sz):
            out[i] = inp[i]
        return sz

    @heavydb('int32(ColumnList<T>, OutputColumn<T>)', T=['int32'], devices=['cpu'])
    def columnlist_set_output_row_size(lst, out):
        set_output_row_size(lst.ncols)
        for j in range(lst.ncols):
            col = lst[j]
            out[j] = 0
            for i in range(lst.nrows):
                out[j] += col[i]
        return len(out)


@pytest.mark.parametrize('T', table_columns_map)
def test_column_alloc(heavydb, T):
    fn = 'column_set_output_row_size'
    col = table_columns_map[T]

    query = (f'select * from table({fn}(cursor('
             f'select {col} from {heavydb.table_name})));')

    _, result = heavydb.sql_execute(query)

    if 'f' in col:
        expected = [(0.0,), (1.0,), (2.0,), (3.0,), (4.0,)]
    else:
        expected = [(0,), (1,), (2,), (3,), (4,)]
    assert list(result) == expected


@pytest.mark.parametrize("fn", ('column_list_row_sum', 'columnlist_set_output_row_size'))
def test_columnlist_alloc(heavydb, fn):

    query = (f'select * from table({fn}(cursor('
             f'select i4, i4+1, i4+2 from {heavydb.table_name})));')

    _, result = heavydb.sql_execute(query)

    assert list(result) == [(10,), (15,), (20,)]



'''RBC String Dictionary Proxy type
'''

__all__ = ['StringDictionaryProxyNumbaType', 'StringDictionaryProxy']


from llvmlite import ir
from numba.core import extending
from numba.core import types as nb_types

from rbc.external import external
from rbc.errors import NumbaTypeError

from . import text_encoding_none
from .metatype import HeavyDBMetaType
from .opaque_pointer import HeavyDBOpaquePtr, OpaquePtrNumbaType
from .utils import as_voidptr

int8_t = ir.IntType(8)
int32_t = ir.IntType(32)


class StringDictionaryProxy(metaclass=HeavyDBMetaType):

    def getStringId(self, str_arg: 'text_encoding_none.TextEncodingNone') -> int:
        """
        """

    def getOrAddTransient(self, str_arg: 'text_encoding_none.TextEncodingNone') -> int:
        """
        """

    def getString(self, index: int) -> 'text_encoding_none.TextEncodingNone':
        """
        """


class HeavyDBStringDictProxyType(HeavyDBOpaquePtr):
    """RowFunctionManager<> is a typesystem custom type that
    represents a class type with the following public interface:

      struct RowFunctionManager { }
    """

    @property
    def numba_type(self):
        return StringDictionaryProxyNumbaType

    @property
    def type_name(self):
        return "StringDictionaryProxy"


class StringDictionaryProxyNumbaType(OpaquePtrNumbaType):
    pass


HeavyDBStringDictProxyType().register_datamodel()


@extending.overload_method(StringDictionaryProxyNumbaType, 'getStringId')
@extending.overload_method(StringDictionaryProxyNumbaType, 'getOrAddTransient')
def ov_getStringId(proxy, str_arg):
    getStringId_ = external("int32 StringDictionaryProxy_getStringId(int8*, int8*)")

    if isinstance(str_arg, nb_types.UnicodeType):
        def impl(proxy, str_arg):
            return getStringId_(as_voidptr(proxy), as_voidptr(str_arg._data))
        return impl
    elif isinstance(str_arg, text_encoding_none.TextEncodingNonePointer):
        def impl(proxy, str_arg):
            return getStringId_(as_voidptr(proxy), as_voidptr(str_arg.ptr))
        return impl
    else:
        raise NumbaTypeError(f'Cannot handle string argument type {str_arg}')


@extending.overload_method(StringDictionaryProxyNumbaType, 'getString')
def heavydb_column_getString(proxy, string_id):
    getBytes = external('int8* StringDictionaryProxy_getStringBytes(int8*, int32)')
    getBytesLength = external('int64 StringDictionaryProxy_getStringLength(int8*, int32)')

    def impl(proxy, string_id):
        ptr = getBytes(as_voidptr(proxy), string_id)
        len = getBytesLength(as_voidptr(proxy), string_id)
        return text_encoding_none.TextEncodingNone(ptr, len)
    return impl



import pytest
import numpy as np
from rbc.tests import heavydb_fixture
from rbc.externals.heavydb import set_output_row_size
from rbc.heavydb import Timestamp

rbc_heavydb = pytest.importorskip('rbc.heavydb')
available_version, reason = rbc_heavydb.is_available()


@pytest.fixture(scope='module')
def heavydb():
    for o in heavydb_fixture(globals(), debug=False,
                             suffices=['timestamp'], minimal_version=(6, 0)):
        define(o)
        yield o


def define(heavydb):

    @heavydb('int32_t(Column<int64_t>, RowMultiplier, OutputColumn<Timestamp>)',
             devices=['cpu', 'gpu'])
    def int_ctor(x, m, y):
        for i in range(len(x)):
            y[i] = Timestamp(x[i])
        return len(x)

    @heavydb('int32_t(Column<int64_t>, RowMultiplier, OutputColumn<Timestamp>)',
             devices=['cpu', 'gpu'])
    def text_ctor(x, m, y):
        for i in range(len(x)):
            if i == 0:
                y[i] = Timestamp("2021-01-01 01:01:01.001001001")
            elif i == 1:
                y[i] = Timestamp("2022-02-02 02:02:02.002002002")
            else:
                y[i] = Timestamp("2023-03-03 03:03:03.003003003")
        return len(x)

    @heavydb('int32_t(Column<Timestamp>, RowMultiplier, OutputColumn<int64_t>)',
             devices=['cpu', 'gpu'])
    def timestamp_null_check(x, m, y):
        for i in range(len(x)):
            y[i] = x.is_null(i)
        return len(x)

    @heavydb('int32_t(Column<Timestamp>, RowMultiplier, OutputColumn<Timestamp>)',
             devices=['cpu', 'gpu'])
    def timestamp_set_null(x, m, y):
        for i in range(len(x)):
            y.set_null(i)
        return len(x)

    @heavydb("int32_t(Column<Timestamp>, TextEncodingNone, OutputColumn<int64_t>)",
             devices=['cpu'])
    def timestamp_get(x, op, y):
        set_output_row_size(len(x))
        for i in range(len(x)):
            if op == 'getYear':
                y[i] = x[i].getYear()
            elif op == 'getMonth':
                y[i] = x[i].getMonth()
            elif op == 'getDay':
                y[i] = x[i].getDay()
            elif op == 'getHours':
                y[i] = x[i].getHours()
            elif op == 'getMinutes':
                y[i] = x[i].getMinutes()
            elif op == 'getSeconds':
                y[i] = x[i].getSeconds()
            elif op == 'getMilliseconds':
                y[i] = x[i].getMilliseconds()
            elif op == 'getMicroseconds':
                y[i] = x[i].getMicroseconds()
            elif op == 'getNanoseconds':
                y[i] = x[i].getNanoseconds()
            else:
                y[i] = -1
        return len(x)

    @heavydb("int32(Column<Timestamp>, TextEncodingNone, OutputColumn<Timestamp>)",
             devices=['cpu'])
    def timestamp_truncate_cpu(x, op, y):
        set_output_row_size(len(x))
        for i in range(len(x)):
            if op == 'truncateToYear':
                y[i] = x[i].truncateToYear()
            elif op == 'truncateToMonth':
                y[i] = x[i].truncateToMonth()
            elif op == 'truncateToDay':
                y[i] = x[i].truncateToDay()
            elif op == 'truncateToHours':
                y[i] = x[i].truncateToHours()
            elif op == 'truncateToMinutes':
                y[i] = x[i].truncateToMinutes()
            elif op == 'truncateToSeconds':
                y[i] = x[i].truncateToSeconds()
            elif op == 'truncateToMilliseconds':
                y[i] = x[i].truncateToMilliseconds()
            elif op == 'truncateToMicroseconds':
                y[i] = x[i].truncateToMicroseconds()
        return len(x)

    @heavydb('int32_t(Column<Timestamp>, Column<Timestamp>, TextEncodingNone, OutputColumn<int32_t>)',  # noqa: E501
             devices=['cpu'])
    def timestamp_operator_cpu(a, b, op, y):
        set_output_row_size(len(a))
        for i in range(len(a)):
            if op == 'eq':
                y[i] = (a[i] == b[i])
            elif op == 'ne':
                y[i] = (a[i] != b[i])
            elif op == 'lt':
                y[i] = (a[i] < b[i])
            elif op == 'le':
                y[i] = (a[i] <= b[i])
            elif op == 'gt':
                y[i] = (a[i] > b[i])
            elif op == 'ge':
                y[i] = (a[i] >= b[i])
            else:
                y[i] = 127
        return len(a)

    @heavydb('int32_t(Column<Timestamp>, Column<Timestamp>, TextEncodingNone, OutputColumn<Timestamp>)',  # noqa: E501
             devices=['cpu'])
    def timestamp_operator_2_cpu(a, b, op, y):
        set_output_row_size(len(a))
        for i in range(len(a)):
            if op == 'floordiv':
                y[i] = (a[i] // b[i])
            elif op == 'truediv':
                y[i] = (a[i] / b[i])
            elif op == 'add':
                y[i] = (a[i] + b[i])
            elif op == 'sub':
                y[i] = (a[i] - b[i])
            elif op == 'mul':
                y[i] = (a[i] * b[i])
            else:
                y[i] = Timestamp(127)
        return len(a)

    if heavydb.has_cuda:
        @heavydb('int32(Column<Timestamp>, RowMultiplier, int32_t, OutputColumn<Timestamp>)',
                 devices=['gpu'])
        def timestamp_truncate_gpu(x, m, op, y):
            for i in range(len(x)):
                if op == 0:
                    y[i] = x[i].truncateToSeconds()
                elif op == 1:
                    y[i] = x[i].truncateToMilliseconds()
                elif op == 2:
                    y[i] = x[i].truncateToMicroseconds()
            return len(x)

        @heavydb('int32_t(Column<Timestamp>, Column<Timestamp>, RowMultiplier, int32_t, OutputColumn<int32_t>)',  # noqa: E501
                 devices=['gpu'])
        def timestamp_operator_gpu(a, b, m, op, y):
            for i in range(len(a)):
                if op == 0:
                    y[i] = (a[i] == b[i])
                elif op == 1:
                    y[i] = (a[i] != b[i])
                elif op == 2:
                    y[i] = (a[i] < b[i])
                elif op == 3:
                    y[i] = (a[i] <= b[i])
                elif op == 4:
                    y[i] = (a[i] > b[i])
                elif op == 5:
                    y[i] = (a[i] >= b[i])
                else:
                    y[i] = 127
            return len(a)

        @heavydb('int32_t(Column<Timestamp>, Column<Timestamp>, RowMultiplier, int32_t, OutputColumn<Timestamp>)',  # noqa: E501
                 devices=['gpu'])
        def timestamp_operator_2_gpu(a, b, m, op, y):
            for i in range(len(a)):
                if op == 0:
                    y[i] = (a[i] + b[i])
                elif op == 1:
                    y[i] = (a[i] - b[i])
                elif op == 2:
                    y[i] = (a[i] * b[i])
                elif op == 3:
                    y[i] = (a[i] // b[i])
                elif op == 4:
                    y[i] = (a[i] / b[i])
                else:
                    y[i] = Timestamp(127)
            return len(a)


@pytest.mark.parametrize("method", ["year", "month", "day", "hours", "minutes", "seconds",
                                    "milliseconds", "microseconds"])
@pytest.mark.parametrize("col", ['t9'])
def test_getMethods(heavydb, col, method):

    table = f"{heavydb.table_name}timestamp"
    query = (f"select * from table(timestamp_get(cursor("
             f"select {col} from {table}), 'get{method.capitalize()}'));")
    _, result = heavydb.sql_execute(query)

    if method == "year":
        assert list(result) == [(1971,), (1972,), (1973,)]
    elif method == "month":
        assert list(result) == [(1,), (2,), (3,)]
    elif method == "day":
        assert list(result) == [(1,), (2,), (3,)]
    elif method == "hours":
        assert list(result) == [(1,), (2,), (3,)]
    elif method == "minutes":
        assert list(result) == [(1,), (2,), (3,)]
    elif method == "seconds":
        assert list(result) == [(1,), (2,), (3,)]
    elif method == "milliseconds":
        assert list(result) == [(1001,), (2002,), (3003,)]
    elif method == "microseconds":
        assert list(result) == [(1001001,), (2002002,), (3003003,)]
    elif method == "nanoseconds":
        assert list(result) == [(1001001001,), (2002002002,), (3003003003,)]
    else:
        assert 0


@pytest.mark.parametrize('device', ['cpu', 'gpu'])
@pytest.mark.parametrize("method", ["year", "month", "day", "hours", "minutes", "seconds",
                                    "milliseconds", "microseconds"])
@pytest.mark.parametrize("col", ['t9'])
def test_truncateMethods(heavydb, device, col, method):

    if device == 'gpu' and not heavydb.has_cuda:
        pytest.skip('test requires CUDA-enabled heavydb server')

    table = f"{heavydb.table_name}timestamp"
    if device == 'cpu':
        query = (f"select * from table(timestamp_truncate_{device}("
                 f"cursor(select {col} from {table}), 'truncateTo{method.capitalize()}'));")
    else:
        methods = ['seconds', 'milliseconds', 'microseconds']
        if method not in methods:
            pytest.skip(f"truncateTo{method.capitalize()}() not implemented for GPU")
        idx = methods.index(method)
        query = (f"select * from table(timestamp_truncate_{device}("
                 f"cursor(select {col} from {table}), 1, {idx}));")
    _, result = heavydb.sql_execute(query)

    if method == "year":
        result = list(map(lambda x: np.datetime64(x[0], 'ns'), result))
        expected = [np.datetime64("1971-01-01 00:00:00.000000000"),
                    np.datetime64("1972-01-01 00:00:00.000000000"),
                    np.datetime64("1973-01-01 00:00:00.000000000")]
        assert result == expected
    elif method == "month":
        result = list(map(lambda x: np.datetime64(x[0], 'ns'), result))
        expected = [np.datetime64("1971-01-01 00:00:00.000000000"),
                    np.datetime64("1972-02-01 00:00:00.000000000"),
                    np.datetime64("1973-03-01 00:00:00.000000000")]
        assert result == expected
    elif method == "day":
        result = list(map(lambda x: np.datetime64(x[0], 'ns'), result))
        expected = [np.datetime64("1971-01-01 00:00:00.000000000"),
                    np.datetime64("1972-02-02 00:00:00.000000000"),
                    np.datetime64("1973-03-03 00:00:00.000000000")]
        assert result == expected
    elif method == "hours":
        result = list(map(lambda x: np.datetime64(x[0], 'ns'), result))
        expected = [np.datetime64("1971-01-01 01:00:00.000000000"),
                    np.datetime64("1972-02-02 02:00:00.000000000"),
                    np.datetime64("1973-03-03 03:00:00.000000000")]
        assert result == expected
    elif method == "minutes":
        result = list(map(lambda x: np.datetime64(x[0], 'ns'), result))
        expected = [np.datetime64("1971-01-01 01:01:00.000000000"),
                    np.datetime64("1972-02-02 02:02:00.000000000"),
                    np.datetime64("1973-03-03 03:03:00.000000000")]
        assert result == expected
    elif method == "seconds":
        result = list(map(lambda x: np.datetime64(x[0], 'ns'), result))
        expected = [np.datetime64("1971-01-01 01:01:01.000000000"),
                    np.datetime64("1972-02-02 02:02:02.000000000"),
                    np.datetime64("1973-03-03 03:03:03.000000000")]
        assert result == expected
    elif method == "milliseconds":
        result = list(map(lambda x: np.datetime64(x[0], 'ns'), result))
        expected = [np.datetime64("1971-01-01 01:01:01.001000000"),
                    np.datetime64("1972-02-02 02:02:02.002000000"),
                    np.datetime64("1973-03-03 03:03:03.003000000")]
        assert result == expected
    elif method == "microseconds":
        result = list(map(lambda x: np.datetime64(x[0], 'ns'), result))
        expected = [np.datetime64("1971-01-01 01:01:01.001001000"),
                    np.datetime64("1972-02-02 02:02:02.002002000"),
                    np.datetime64("1973-03-03 03:03:03.003003000")]
        assert result == expected
    else:
        assert 0


@pytest.mark.parametrize('device', ['cpu', 'gpu'])
@pytest.mark.parametrize('op', ['eq', 'ne', 'lt', 'le', 'gt', 'ge'])
@pytest.mark.parametrize('col1', ['t9', 't9_2'])
@pytest.mark.parametrize('col2', ['t9_2', 't9'])
def test_operators(heavydb, device, col1, col2, op):

    if device == 'gpu' and not heavydb.has_cuda:
        pytest.skip('test requires CUDA-enabled heavydb server')

    table_op = {
        'eq': '=',
        'ne': '!=',
        'lt': '<',
        'le': '<=',
        'gt': '>',
        'ge': '>=',

    }

    table = f'{heavydb.table_name}timestamp'
    _, expected = heavydb.sql_execute(f'select {col1} {table_op[op]} {col2} from {table}')

    if device == 'cpu':
        query = (f"select * from table(timestamp_operator_{device}("
                 f"cursor(select {col1} from {table}),"
                 f"cursor(select {col2} from {table}),"
                 f"'{op}'))")
    else:
        idx = list(table_op.keys()).index(op)
        query = (f"select * from table(timestamp_operator_{device}("
                 f"cursor(select {col1} from {table}),"
                 f"cursor(select {col2} from {table}),"
                 f"{idx}))")
    _, result = heavydb.sql_execute(query)

    assert list(result) == list(expected)


@pytest.mark.parametrize('device', ['cpu', 'gpu'])
@pytest.mark.parametrize('op', ['add', 'sub', 'mul', 'floordiv', 'truediv'])
def test_operators2(heavydb, device, op):

    if device == 'gpu' and not heavydb.has_cuda:
        pytest.skip('test requires CUDA-enabled heavydb server')

    col1 = 't9_2'
    col2 = 't9'

    table = f'{heavydb.table_name}timestamp'
    if device == 'cpu':
        query = (f"select * from table(timestamp_operator_2_{device}("
                 f"cursor(select {col1} from {table}),"
                 f"cursor(select {col2} from {table}),"
                 f"'{op}'))")
    else:
        idx = ['add', 'sub', 'mul', 'floordiv', 'truediv'].index(op)
        query = (f"select * from table(timestamp_operator_2_{device}("
                 f"cursor(select {col1} from {table}),"
                 f"cursor(select {col2} from {table}),"
                 f"1,"
                 f"{idx}));")
    _, result = heavydb.sql_execute(query)

    if op == 'add':
        result = list(map(lambda x: np.datetime64(x[0], 'ns'), result))
        expected = [np.datetime64('2022-01-01 02:02:02.002002002'),
                    np.datetime64('2024-03-05 04:04:04.004004004'),
                    np.datetime64('2026-05-03 06:06:06.006006006')]
        assert result == expected
    elif op == 'sub':
        result = list(map(lambda x: np.datetime64(x[0], 'ns'), result))
        expected = [np.datetime64('2020-01-02'),
                    np.datetime64('2020-01-02'),
                    np.datetime64('2020-01-01')]
        assert result == expected
    elif op == 'mul':
        result = list(map(lambda x: np.datetime64(x[0], 'ns'), result))
        expected = [np.datetime64('2185-09-18 11:54:21.428053649'),
                    np.datetime64('1732-10-05 06:40:02.933971524'),
                    np.datetime64('2139-05-27 07:08:53.212359449')]
        assert result == expected
    elif op == 'floordiv':
        result = list(map(lambda x: np.datetime64(x[0], 'ns'), result))
        expected = [np.datetime64('1970-01-01 00:00:00.000000051'),
                    np.datetime64('1970-01-01 00:00:00.000000024'),
                    np.datetime64('1970-01-01 00:00:00.000000016')]
        assert result == expected
    elif op == 'truediv':
        result = list(map(lambda x: np.datetime64(x[0], 'ns'), result))
        expected = [np.datetime64('1970-01-01 00:00:00.000000051'),
                    np.datetime64('1970-01-01 00:00:00.000000024'),
                    np.datetime64('1970-01-01 00:00:00.000000016')]
        assert result == expected
    else:
        assert 0


@pytest.mark.parametrize('col', ['i8_2'])
def test_int_ctor(heavydb, col):
    table = f"{heavydb.table_name}timestamp"

    query = (f"select * from table(int_ctor("
             f"cursor(select {col} from {table}), 1))")
    _, result = heavydb.sql_execute(query)
    _, expected = heavydb.sql_execute(f"select t9_2 from {table}")

    assert list(result) == list(expected)


def test_literal_string_ctor(heavydb):
    table = f"{heavydb.table_name}timestamp"

    query = (f"select * from table(text_ctor("
             f"cursor(select i8_2 from {table})))")
    _, result = heavydb.sql_execute(query)
    _, expected = heavydb.sql_execute(f"select t9_2 from {table}")

    assert list(result) == list(expected)


@pytest.mark.parametrize('col', ['t9_null'])
def test_is_null(heavydb, col):
    table = f"{heavydb.table_name}timestamp"

    query = (f"select * from table(timestamp_null_check("
             f"cursor(select {col} from {table})))")
    _, result = heavydb.sql_execute(query)

    assert list(result) == [(0,), (1,), (0,)]


@pytest.mark.parametrize('col', ['t9'])
def test_set_null(heavydb, col):
    table = f"{heavydb.table_name}timestamp"

    query = (f"select * from table(timestamp_set_null("
             f"cursor(select {col} from {table})))")
    _, result = heavydb.sql_execute(query)

    assert list(result) == [(None,), (None,), (None,)]



"""
Array API specification for statistical functions.

https://data-apis.org/array-api/latest/API_specification/statistical_functions.html
"""
from rbc.externals.stdio import printf
from rbc import typesystem
from rbc.heavydb import ArrayPointer
from rbc.stdlib import Expose
from numba.core import extending, types, errors
from numba.np import numpy_support
import numpy as np


__all__ = [
    'min', 'max', 'mean', 'prod', 'sum', 'std', 'var'
]


expose = Expose(globals(), 'statistical_functions')


def _get_type_limits(eltype):
    np_dtype = numpy_support.as_dtype(eltype)
    if isinstance(eltype, types.Integer):
        return np.iinfo(np_dtype)
    elif isinstance(eltype, types.Float):
        return np.finfo(np_dtype)
    else:
        msg = 'Type {} not supported'.format(eltype)
        raise errors.TypingError(msg)


@extending.overload(max)
@expose.implements('max')
@extending.overload_method(ArrayPointer, 'max')
def _impl_array_max(x):
    """
    Calculates the maximum value of the input array x
    """
    if isinstance(x, ArrayPointer):
        # the array api standard says this is implementation specific
        limits = _get_type_limits(x.eltype)
        t = typesystem.Type.fromobject(x.eltype)
        if t.is_float:
            min_value = limits.min
        elif t.is_int or t.is_uint:
            min_value = 0 if t.is_uint else limits.min + 1
        else:
            raise TypeError(f'Unsupported type {t}')

        def impl(x):
            if len(x) <= 0:
                printf("impl_array_max: cannot find max of zero-sized array")  # noqa: E501
                return min_value
            m = x[0]
            for i in range(len(x)):
                v = x[i]
                if v > m:
                    m = v
            return m
        return impl


@extending.overload(min)
@expose.implements('min')
@extending.overload_method(ArrayPointer, 'min')
def _impl_array_min(x):
    """
    Calculates the minimum value of the input array x.
    """
    if isinstance(x, ArrayPointer):
        max_value = _get_type_limits(x.eltype).max

        def impl(x):
            if len(x) <= 0:
                printf("impl_array_min: cannot find min of zero-sized array")  # noqa: E501
                return max_value
            m = x[0]
            for i in range(len(x)):
                v = x[i]
                if v < m:
                    m = v
            return m
        return impl


@extending.overload(sum)
@expose.implements('sum')
@extending.overload_method(ArrayPointer, 'sum')
def _impl_np_sum(a):
    """
    Calculates the sum of the input array x.
    """
    if isinstance(a, ArrayPointer):
        def impl(a):
            s = 0
            n = len(a)
            for i in range(n):
                s += a[i]
            return s
        return impl


@expose.implements('prod')
@extending.overload_method(ArrayPointer, 'prod')
def _impl_np_prod(a):
    """
    Calculates the product of input array x elements.
    """
    if isinstance(a, ArrayPointer):
        def impl(a):
            s = 1
            n = len(a)
            for i in range(n):
                s *= a[i]
            return s
        return impl


@expose.implements('mean')
@extending.overload_method(ArrayPointer, 'mean')
def _impl_array_mean(x):
    """
    Calculates the arithmetic mean of the input array x.
    """
    zero_value = np.nan

    if isinstance(x, ArrayPointer):
        def impl(x):
            if len(x) == 0:
                printf("Mean of empty array\n")
                return zero_value
            return sum(x) / len(x)
        return impl


@expose.not_implemented('std')
def _impl_array_std(x, axis=None, correction=0.0, keepdims=False):
    """
    Calculates the standard deviation of the input array x.
    """
    pass


@expose.not_implemented('var')
def _impl_array_var(x, axis=None, correction=0.0, keepdims=False):
    """
    Calculates the variance of the input array x.
    """
    pass



"""RBC Timestamp type that corresponds to HeavyDB type timestamp."""
import operator
import numpy as np
from rbc import typesystem
from rbc.heavydb import HeavyDBMetaType
from rbc.external import external
from rbc.targetinfo import TargetInfo
from numba.core import extending, cgutils, datamodel
from numba.core import types as nb_types
from llvmlite import ir
from typing import Union

__all__ = ['HeavyDBTimestampType', 'Timestamp', 'TimestampNumbaType']


kNanoSecsPerSec = int(np.timedelta64(1, 's') / np.timedelta64(1, 'ns'))
kMicroSecsPerSec = int(np.timedelta64(1, 's') / np.timedelta64(1, 'us'))
kMilliSecsPerSec = int(np.timedelta64(1, 's') / np.timedelta64(1, 'ms'))
kSecsPerMin = 60
kMinsPerHour = 60
kHoursPerDay = 24
kSecsPerHour = int(np.timedelta64(1, 'h') / np.timedelta64(1, 's'))
kSecsPerDay = int(np.timedelta64(1, 'D') / np.timedelta64(1, 's'))
kDaysPerWeek = 7
kMonsPerYear = 12


class Timestamp(metaclass=HeavyDBMetaType):
    """
    RBC ``Timestamp`` type that corresponds to HeavyDB type TIMESTAMP.

    .. code-block:: c

        struct Timestamp {
            int64_t time;
        };

    .. code-block:: python

        from rbc.heavydb import Timestamp
        @heavydb('int32_t(Column<Timestamp>, RowMultiplier, Column<int64_t>')
        def get_years(column_times, m, column_hours):
            for i in range(len(column_times)):
                column_hours[i] = column_times[i].getHours()
            return len(column_times)

    All comparison (dunder methods) are implemented for this class.
    """

    def __init__(self, time: Union[int, 'nb_types.StringLiteral']) -> 'Timestamp':
        pass

    def truncateToMicroseconds(self) -> 'Timestamp':
        """
        Return a new ``Timestamp`` with ``time`` truncated to microseconds
        """
        pass

    def truncateToMilliseconds(self) -> 'Timestamp':
        """
        Return a new ``Timestamp`` with ``time`` truncated to milliseconds
        """
        pass

    def truncateToSeconds(self) -> 'Timestamp':
        """
        Return a new ``Timestamp`` with ``time`` truncated to seconds
        """
        pass

    def truncateToMinutes(self) -> 'Timestamp':
        """
        Return a new ``Timestamp`` with ``time`` truncated to minutes

        .. note::
            Only available on ``CPU``
        """
        pass

    def truncateToHours(self) -> 'Timestamp':
        """
        Return a new ``Timestamp`` with ``time`` truncated to hours

        .. note::
            Only available on ``CPU``
        """
        pass

    def truncateToDay(self) -> 'Timestamp':
        """
        Return a new ``Timestamp`` with ``time`` truncated to day

        .. note::
            Only available on ``CPU``
        """
        pass

    def truncateToMonth(self) -> 'Timestamp':
        """
        Return a new ``Timestamp`` with ``time`` truncated to month

        .. note::
            Only available on ``CPU``
        """
        pass

    def truncateToYear(self) -> 'Timestamp':
        """
        Return a new ``Timestamp`` with ``time`` truncated to year

        .. note::
            Only available on ``CPU``
        """
        pass

    def getMicroseconds(self) -> int:
        """
        Get ``time`` as microseconds

        .. note::
            Only available on ``CPU``
        """
        pass

    def getMilliseconds(self) -> int:
        """
        Get ``time`` as milliseconds

        .. note::
            Only available on ``CPU``
        """
        pass

    def getSeconds(self) -> int:
        """
        Get ``time`` as seconds

        .. note::
            Only available on ``CPU``
        """
        pass

    def getMinutes(self) -> int:
        """
        Get ``time`` as minutes

        .. note::
            Only available on ``CPU``
        """
        pass

    def getHours(self) -> int:
        """
        Get ``time`` as hours

        .. note::
            Only available on ``CPU``
        """
        pass

    def getDay(self) -> int:
        """
        Get ``time`` as day

        .. note::
            Only available on ``CPU``
        """
        pass

    def getMonth(self) -> int:
        """
        Get ``time`` as month

        .. note::
            Only available on ``CPU``
        """
        pass

    def getYear(self) -> int:
        """
        Get ``time`` as year

        .. note::
            Only available on ``CPU``
        """
        pass

    def __eq__(self, other: 'Timestamp') -> bool:
        """
        """
        pass

    def __ne__(self, other: 'Timestamp') -> bool:
        """
        """
        pass

    def __lt__(self, other: 'Timestamp') -> bool:
        """
        """
        pass

    def __le__(self, other: 'Timestamp') -> bool:
        """
        """
        pass

    def __gt__(self, other: 'Timestamp') -> bool:
        """
        """
        pass

    def __ge__(self, other: 'Timestamp') -> bool:
        """
        """
        pass

    def __add__(self, other: 'Timestamp') -> 'Timestamp':
        """
        """
        pass

    def __sub__(self, other: 'Timestamp') -> 'Timestamp':
        """
        """
        pass

    def __mul__(self, other: 'Timestamp') -> 'Timestamp':
        """
        """
        pass

    def __floordiv__(self, other: 'Timestamp') -> 'Timestamp':
        """
        """
        pass

    def __truediv__(self, other: 'Timestamp') -> 'Timestamp':
        """
        """
        pass


class TimestampNumbaType(nb_types.Type):
    bitwidth = 64

    def __init__(self):
        super().__init__(name='Timestamp')


@extending.register_model(TimestampNumbaType)
class IntervalModel(datamodel.models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [
            ('time', nb_types.int64),
        ]
        datamodel.models.StructModel.__init__(self, dmm, fe_type, members)


extending.make_attribute_wrapper(TimestampNumbaType, 'time', 'time')


class HeavyDBTimestampType(typesystem.Type):
    """Typesystem type class for HeavyDB timestamp structures.
    """
    @property
    def __typesystem_type__(self):
        return typesystem.Type.fromstring('{int64 time}').params(
            name='Timestamp', NumbaType=TimestampNumbaType)

    def tonumba(self, bool_is_int8=None):
        return TimestampNumbaType()

    def postprocess_type(self):
        return self.params(shorttypename='Timestamp')


@extending.type_callable(Timestamp)
def type_heavydb_timestamp(context):
    def typer(arg):
        if isinstance(arg, (nb_types.Integer, nb_types.StringLiteral)):
            return typesystem.Type.fromobject('Timestamp').tonumba()

    return typer


@extending.lower_builtin(Timestamp, nb_types.Integer)
def heavydb_timestamp_int_ctor(context, builder, sig, args):
    time = args[0]
    typ = sig.return_type
    timestamp = cgutils.create_struct_proxy(typ)(context, builder)
    timestamp.time = time
    return timestamp._getvalue()


@extending.lower_builtin(Timestamp, nb_types.StringLiteral)
def heavydb_timestamp_literal_string_ctor(context, builder, sig, args):
    i64 = ir.IntType(64)
    time = i64(np.datetime64(sig.args[0].literal_value).astype('long'))
    typ = sig.return_type
    timestamp = cgutils.create_struct_proxy(typ)(context, builder)
    timestamp.time = time
    return timestamp._getvalue()


@extending.lower_cast(nb_types.Integer, TimestampNumbaType)
def cast_int_to_timestamp(context, builder, fromty, toty, val):
    timestamp = cgutils.create_struct_proxy(toty)(context, builder)
    timestamp.time = val
    return timestamp._getvalue()


@extending.overload_method(TimestampNumbaType, "getYear")
def heavydb_timestamp_getYear(timestamp):
    target_info = TargetInfo()
    if target_info.is_cpu:
        extract_year = external('int64 extract_year(int64)')

        def impl(timestamp):
            return extract_year(timestamp.time // kNanoSecsPerSec)
        return impl


@extending.overload_method(TimestampNumbaType, "getMonth")
def heavydb_timestamp_getMonth(timestamp):
    target_info = TargetInfo()
    if target_info.is_cpu:
        extract_month = external('int64 extract_month(int64)')

        def impl(timestamp):
            return extract_month(timestamp.time // kNanoSecsPerSec)
        return impl


@extending.overload_method(TimestampNumbaType, "getDay")
def heavydb_timestamp_getDay(timestamp):
    target_info = TargetInfo()
    if target_info.is_cpu:
        extract_day = external('int64 extract_day(int64)')

        def impl(timestamp):
            return extract_day(timestamp.time // kNanoSecsPerSec)
        return impl


@extending.overload_method(TimestampNumbaType, "getHours")
def heavydb_timestamp_getHours(timestamp):
    target_info = TargetInfo()
    if target_info.is_cpu:
        extract_hour = external('int64 extract_hour(int64)')

        def impl(timestamp):
            return extract_hour(timestamp.time // kNanoSecsPerSec)
        return impl


@extending.overload_method(TimestampNumbaType, "getMinutes")
def heavydb_timestamp_getMinutes(timestamp):
    target_info = TargetInfo()
    if target_info.is_cpu:
        extract_minute = external('int64 extract_minute(int64)')

        def impl(timestamp):
            return extract_minute(timestamp.time // kNanoSecsPerSec)
        return impl


@extending.overload_method(TimestampNumbaType, "getSeconds")
def heavydb_timestamp_getSeconds(timestamp):
    target_info = TargetInfo()
    if target_info.is_cpu:
        extract_second = external('int64 extract_second(int64)')

        def impl(timestamp):
            return extract_second(timestamp.time // kNanoSecsPerSec)
        return impl


@extending.overload_method(TimestampNumbaType, "getMilliseconds")
def heavydb_timestamp_getMilliseconds(timestamp):
    target_info = TargetInfo()
    if target_info.is_cpu:
        extract_millisecond = external('int64 extract_millisecond(int64)')

        def impl(timestamp):
            return extract_millisecond(
                timestamp.time // (kNanoSecsPerSec // kMilliSecsPerSec))
        return impl


@extending.overload_method(TimestampNumbaType, "getMicroseconds")
def heavydb_timestamp_getMicroseconds(timestamp):
    target_info = TargetInfo()
    if target_info.is_cpu:
        extract_microsecond = external('int64 extract_microsecond(int64)')

        def impl(timestamp):
            return extract_microsecond(
                timestamp.time // (kNanoSecsPerSec // kMicroSecsPerSec))
        return impl


@extending.overload_method(TimestampNumbaType, "getNanoseconds")
def heavydb_timestamp_getNanoseconds(timestamp):
    target_info = TargetInfo()
    if target_info.is_cpu:
        extract_nanosecond = external('int64 extract_nanosecond(int64)')

        def impl(timestamp):
            return extract_nanosecond(timestamp.time)
        return impl


@extending.overload_method(TimestampNumbaType, "truncateToYear")
def heavydb_timestamp_truncateToYear(timestamp):
    target_info = TargetInfo()
    if target_info.is_cpu:
        datetrunc_year = external('int64 datetrunc_year(int64)')

        def impl(timestamp):
            timeval = datetrunc_year(timestamp.time // kNanoSecsPerSec) * kNanoSecsPerSec
            return Timestamp(timeval)
        return impl


@extending.overload_method(TimestampNumbaType, "truncateToMonth")
def heavydb_timestamp_truncateToMonth(timestamp):
    target_info = TargetInfo()
    if target_info.is_cpu:
        datetrunc_month = external('int64 datetrunc_month(int64)')

        def impl(timestamp):
            timeval = datetrunc_month(timestamp.time // kNanoSecsPerSec) * kNanoSecsPerSec
            return Timestamp(timeval)
        return impl


@extending.overload_method(TimestampNumbaType, "truncateToDay")
def heavydb_timestamp_truncateToDay(timestamp):
    target_info = TargetInfo()
    if target_info.is_cpu:
        datetrunc_day = external('int64 datetrunc_day(int64)')

        def impl(timestamp):
            timeval = datetrunc_day(timestamp.time // kNanoSecsPerSec) * kNanoSecsPerSec
            return Timestamp(timeval)
        return impl


@extending.overload_method(TimestampNumbaType, "truncateToHours")
def heavydb_timestamp_truncateToHour(timestamp):
    target_info = TargetInfo()
    if target_info.is_cpu:
        datetrunc_hour = external('int64 datetrunc_hour(int64)')

        def impl(timestamp):
            timeval = datetrunc_hour(timestamp.time // kNanoSecsPerSec) * kNanoSecsPerSec
            return Timestamp(timeval)
        return impl


@extending.overload_method(TimestampNumbaType, "truncateToMinutes")
def heavydb_timestamp_truncateToMinutes(timestamp):
    target_info = TargetInfo()
    if target_info.is_cpu:
        datetrunc_minute = external('int64 datetrunc_minute(int64)')

        def impl(timestamp):
            timeval = datetrunc_minute(timestamp.time // kNanoSecsPerSec) * kNanoSecsPerSec
            return Timestamp(timeval)
        return impl


@extending.overload_method(TimestampNumbaType, "truncateToSeconds")
def heavydb_timestamp_truncateToSeconds(timestamp):

    def impl(timestamp):
        timeval = (timestamp.time // kNanoSecsPerSec) * kNanoSecsPerSec
        return Timestamp(timeval)
    return impl


@extending.overload_method(TimestampNumbaType, "truncateToMilliseconds")
def heavydb_timestamp_truncateToMilliseconds(timestamp):

    def impl(timestamp):
        timeval = (timestamp.time // kMicroSecsPerSec) * kMicroSecsPerSec
        return Timestamp(timeval)
    return impl


@extending.overload_method(TimestampNumbaType, "truncateToMicroseconds")
def heavydb_timestamp_truncateToMicroseconds(timestamp):

    def impl(timestamp):
        timeval = (timestamp.time // kMilliSecsPerSec) * kMilliSecsPerSec
        return Timestamp(timeval)
    return impl


def overload_binary_cmp_op(op, retty):

    def heavydb_operator_impl(a, b):
        if isinstance(a, TimestampNumbaType) and isinstance(b, TimestampNumbaType):
            def impl(a, b):
                return retty(nb_types.int64(op(a.time, b.time)))
            return impl

    decorate = extending.overload(op)

    def wrapper(overload_func):
        return decorate(heavydb_operator_impl)

    return wrapper


@overload_binary_cmp_op(operator.eq, bool)
@overload_binary_cmp_op(operator.ne, bool)
@overload_binary_cmp_op(operator.lt, bool)
@overload_binary_cmp_op(operator.le, bool)
@overload_binary_cmp_op(operator.gt, bool)
@overload_binary_cmp_op(operator.ge, bool)
@overload_binary_cmp_op(operator.truediv, Timestamp)
@overload_binary_cmp_op(operator.floordiv, Timestamp)
@overload_binary_cmp_op(operator.add, Timestamp)
@overload_binary_cmp_op(operator.sub, Timestamp)
@overload_binary_cmp_op(operator.mul, Timestamp)
def heavydb_timestamp(a, b):
    pass



import pytest
from rbc.tests import heavydb_fixture


@pytest.fixture(scope='module')
def heavydb():

    for o in heavydb_fixture(globals(), minimal_version=(5, 6), suffices=['']):
        define(o)
        yield o


def define(heavydb):

    @heavydb('int32(Column<int64>, RowMultiplier, OutputColumn<int64>)')
    def text_rbc_copy_c_rowmul(x, m, y):
        for i in range(len(x)):
            # TODO: implement is_null support
            y[i] = x[i]
        return len(x)

    @heavydb('int32(Cursor<int64, double>, RowMultiplier,'
             ' OutputColumn<int64>, OutputColumn<double>)')
    def text_rbc_copy_cc_rowmul(x, x2, m, y, y2):
        for i in range(len(x)):
            y[i] = x[i]
            y2[i] = x2[i]
        return len(x)

    @heavydb('int32(Cursor<int64>, Cursor<double>, RowMultiplier,'
             ' OutputColumn<int64>, OutputColumn<double>)')
    def text_rbc_copy_c_c_rowmul(x, x2, m, y, y2):
        for i in range(len(x)):
            y[i] = x[i]
            y2[i] = x2[i]
        return len(x)

    @heavydb('int32(Cursor<int32>, Cursor<double, int64>, RowMultiplier,'
             ' OutputColumn<int32>, OutputColumn<double>, OutputColumn<int64>)')
    def text_rbc_copy_c_cc_rowmul(x, x2, x3, m, y, y2, y3):
        for i in range(len(x)):
            y[i] = x[i]
            y2[i] = x2[i]
            y3[i] = x3[i]
        return len(x)

    @heavydb('int32(Cursor<int32, int64>, Cursor<double>, RowMultiplier,'
             ' OutputColumn<int32>, OutputColumn<int64>, OutputColumn<double>)')
    def text_rbc_copy_cc_c_rowmul(x, x2, x3, m, y, y2, y3):
        for i in range(len(x)):
            y[i] = x[i]
            y2[i] = x2[i]
            y3[i] = x3[i]
        return len(x)

    @heavydb('int32(Cursor<int32, int64>, Cursor<double, float>, RowMultiplier,'
             ' OutputColumn<int32>, OutputColumn<int64>, OutputColumn<double>,'
             ' OutputColumn<float>)')
    def text_rbc_copy_cc_cc_rowmul(x, x2, x3, x4, m, y, y2, y3, y4):
        for i in range(len(x)):
            y[i] = x[i]
            y2[i] = x2[i]
            y3[i] = x3[i]
            y4[i] = x4[i]
        return len(x)

    @heavydb('int32(Cursor<int32, int64, double, float>, RowMultiplier,'
             ' OutputColumn<int32>, OutputColumn<int64>, OutputColumn<double>,'
             ' OutputColumn<float>)')
    def text_rbc_copy_cccc_rowmul(x, x2, x3, x4, m, y, y2, y3, y4):
        for i in range(len(x)):
            y[i] = x[i]
            y2[i] = x2[i]
            y3[i] = x3[i]
            y4[i] = x4[i]
        return len(x)


@pytest.mark.parametrize('use_default', [True, False])
@pytest.mark.parametrize("inputs",
                         ['i8', 'cursor(i8,f8)', 'i8;f8', 'i4;cursor(f8,i8)',
                          'cursor(i4,i8);f8', 'cursor(i4,i8);cursor(f8,f4)',
                          'cursor(i4,i8,f8,f4)'])
def test_copy(heavydb, use_default, inputs):
    if use_default:
        heavydb.require_version((5, 7), 'Requires heavydb-internal PR 5403')

    groups = inputs.split(';')
    table_names = [f'{heavydb.table_name}'] * len(groups)

    group_results = {}
    args = []
    cc = []
    expected = [()] * 5
    for colnames, table_name in zip(groups, table_names):
        if colnames.startswith('cursor('):
            colnames = colnames[7:-1]
        result = group_results.get(colnames)
        if result is None:
            _, result = heavydb.sql_execute(f'select {colnames} from {table_name}')
            result = group_results[colnames] = list(result)
        expected = [row1 + row2 for row1, row2 in zip(expected, result)]
        args.append(f'cursor(select {colnames} from {table_name})')
        cc.append((colnames.count(',') + 1) * 'c')
    if not use_default:
        args.append('1')
    args = ', '.join(args)
    cc = '_'.join(cc)

    query = f'select * from table(text_rbc_copy_{cc}_rowmul({args}))'
    _, result = heavydb.sql_execute(query)
    result = list(result)

    assert result == expected


@pytest.mark.parametrize('kind', ['1', '11', '111', '211', '221', '212', '13', '32', '34', '242'])
def test_ct_binding_constant_sizer(heavydb, kind):
    heavydb.require_version((5, 7), 'Requires heavydb-internal PR 5274')
    colnames = ', '.join({'1': 'i4', '2': 'i8', '3': 'i4, i4, i4',
                          '4': 'i8, i8, i8'}[n] for n in kind)

    query = (f'select * from table(ct_binding_udtf_constant(cursor('
             f'select {colnames} from {heavydb.table_name})))')
    _, result = heavydb.sql_execute(query)
    result = list(result)

    assert result == [(int(kind),)]


@pytest.mark.parametrize('use_default', [True, False])
@pytest.mark.parametrize('kind', ['19', '119', '1119', '2119', '2219',
                                  '2129', '139', '329', '349', '2429',
                                  '91', '196', '396', '369', '169'])
def test_ct_binding_row_multiplier(heavydb, use_default, kind):
    if use_default:
        heavydb.require_version((5, 7), 'Requires heavydb-internal PR 5403')

    if heavydb.version < (5, 7):
        suffix = {'91': '2', '369': '2', '169': '3'}.get(kind, '')
        codes = {'1': 'i4', '2': 'i8', '3': 'i4, i4, i4', '4': 'i8, i8, i8',
                 '9': '1', '6': 'cast(123 as int)'}
        first = []
        last = []
        cursor = []
        for n in kind:
            if n in '69':
                if cursor:
                    last.append(codes[n])
                else:
                    first.append(codes[n])
            else:
                assert n in '1234'
                cursor.append(codes[n])
        first = ', '.join(first + [''])
        last = ', '.join([''] + last)
        cursor = ', '.join(cursor)
        query = (f'select * from table(ct_binding_udtf{suffix}({first}cursor('
                 f'select {cursor} from {heavydb.table_name}){last}))')
        _, result = heavydb.sql_execute(query)
        result = list(result)

        assert result == [(1000 + int(kind),)], (result, query)
    else:
        suffix = {'91': '2', '369': '5', '169': '3', '196': '6', '396': '4'}.get(kind, '')
        codes = {'1': 'i4', '2': 'i8', '3': 'i4, i4, i4', '4': 'i8, i8, i8',
                 '9': '1', '6': 'cast(123 as int)'}
        input_val = 123 if suffix and suffix in '6543' else 0
        first = []
        last = []
        cursor = []
        for n in kind:
            if n == '6':
                if cursor:
                    last.append(codes[n])
                else:
                    first.append(codes[n])
            elif n == '9':
                if use_default:
                    continue
                if cursor:
                    last.append(codes[n])
                else:
                    first.append(codes[n])
            else:
                assert n in '1234'
                cursor.append(codes[n])
        first = ', '.join(first + [''])
        last = ', '.join([''] + last)
        cursor = ', '.join(cursor)
        query = (f'select * from table(ct_binding_udtf{suffix}({first}cursor('
                 f'select {cursor} from {heavydb.table_name}){last}))')
        _, result = heavydb.sql_execute(query)
        result = list(result)

        assert result == [(1000 + int(kind) + 1 + 10*input_val,)], (result, query)
